//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//This part is from herbert Mod, modified for MorTillo Mod

#include "StdAfx.h"
#include "Amuc.h"
#include "emule.h"
#include "preferences.h"
#include "math.h"
#include "iphlpapi.h"

#pragma comment(lib, "snmpapi.lib" )

HANDLE hPollForTrapEvent = NULL;
AsnObjectIdentifier SupportedView = {0,0};

IMPLEMENT_DYNCREATE(CAmucThread, CWinThread)

CAmucThread::CAmucThread(void){
	m_fpSnmpUtilMemAlloc	= NULL;
	m_fpSnmpUtilMemFree		= NULL;
	m_pvarBindList			= NULL;
	m_bUse_iphlpapi			= false;
	m_hInstIpHlp			= 0;
	isAmuc					= false;
	librariesloaded			= false;
	g_nAveragingWindow		= 20;
	estupload				= 0;
	isEnabled				= false;
	clientsready			= false;
}

CAmucThread::~CAmucThread(void){
	// Close the echo request file handle
	if(librariesloaded){
		isAmuc = false;
		ResumeThread();
		// Wait till AMUC Thread has ended
		while(!isAmuc)
			::Sleep(100);
   	}

	if( m_pvarBindList )
		SnmpUtilMemFree( m_pvarBindList );

    if(m_hInst)
        FreeLibrary(m_hInst);

    if(m_hInstIpHlp)
        FreeLibrary(m_hInstIpHlp);

	m_hInstIpHlp = 0;
	m_hInst = 0;

	librariesloaded = false;
}

int CAmucThread::Run(){
	float		upload;			// upload calculated next
	float		updatarate;		// Datarate of the current eMule upload
	float		otherUpload;	// Datarate of the current upload of other applications
	float		downUpload;	    // Max. allowed upload considering the download

	uint8		nextview=0;

	DWORD		r=0,s=0,elapsed;
	DWORD		dwRecv_bps, dwSent_bps;
    DWORD		dwTime;
    float		dbRecv, dbSent;
    static float dbRecvWrap = 0;
    static float dbSentWrap = 0;

	// Init first values

	GetSentAndReceivedOctets( &r, &s );
	m_dbTotalBytesRecv = (float)r;
    m_dbTotalBytesSent = (float)s;
	m_dwStartTime = GetTickCount();
	m_nArrayIndex = 0;
	upload = 0;

	// Maximal allow Upload Rate thinking of enough bandwith for further download activities
	downUpload = 0.5; //((float)RecvStats[m_nArrayIndex].ave/1024+5)*0.03 +1.0;

	// Over how much seconds should be averaged
	// g_nAveragingWindow = 30;         // Already set

	isAmuc = true;

	while(isAmuc && theApp.emuledlg->IsRunning()){
		if (!theApp.emuledlg->IsRunning()) break; 
		if(theApp.glob_prefs->IsAMUCEnabled() && theApp.glob_prefs->IsAutoUpEnabled()){
			isEnabled = true;

			GetSentAndReceivedOctets(&r, &s);
			dwTime = ::GetTickCount();
			elapsed = dwTime - m_dwStartTime;
			m_dwStartTime = dwTime;

			dbRecv = r + dbRecvWrap;
			dbSent = s + dbSentWrap;




				//check for integer wrap
				if( dbRecv < m_dbTotalBytesRecv )
				{
					dbRecvWrap += 0xffffffff;
					dbRecv = r + dbRecvWrap;
				}

				if( dbSent < m_dbTotalBytesSent )
				{
					dbSentWrap += 0xffffffff;
					dbSent = s + dbSentWrap;
				}

				//total bytes sent or received during this interval
				DWORD total_recv = (DWORD)( dbRecv - m_dbTotalBytesRecv );
				DWORD total_sent = (DWORD)( dbSent - m_dbTotalBytesSent );
				dwRecv_bps = dwSent_bps = 0;

				//calc bits per second
				if( elapsed )
				{
					dwRecv_bps = MulDiv( total_recv, 1000, elapsed );
				    dwSent_bps = MulDiv( total_sent, 1000, elapsed );
				}
				//convert to float
				float recv_bits = (float)(r);
				float sent_bits = (float)(s);

				if(recv_bits < 1.0 || sent_bits <1.0)
					CheckNT();

				//calc the average bps
				CalcAverages( recv_bits, dwTime, dwRecv_bps, &RecvStats[0] );
			    CalcAverages( sent_bits, dwTime, dwSent_bps, &SentStats[0] );

				//save the totals
				m_dbTotalBytesRecv = dbRecv;
			    m_dbTotalBytesSent = dbSent;


			// Current Upload Rate
			updatarate			= (float)theApp.uploadqueue->GetDatarate()/1024;

			downUpload = (float)theApp.downloadqueue->GetDatarate()/1024*0.05;			// Current Upload generated by other applications (not eMule)
			
			otherUpload = (float)SentStats[m_nArrayIndex].ave/1024-updatarate;
			if(otherUpload < ((float)RecvStats[m_nArrayIndex].ave/1024+10)*0.04)
				otherUpload = ((float)RecvStats[m_nArrayIndex].ave/1024+10)*0.04;
			if(otherUpload < ((float)theApp.downloadqueue->GetDatarate()/1024+10)*0.04)
				otherUpload = ((float)theApp.downloadqueue->GetDatarate()/1024+10)*0.04;

			if(otherUpload<0.0)
				otherUpload = 0;


			// Upload should be max. allowed upload minus upload already used by other applications
			upload = theApp.glob_prefs->GetMaxAutoUpload() -downUpload - otherUpload;


			// Set the maximal Upload for eMule
			if(upload>updatarate+5.0)
				estupload = updatarate+5.0;
			else if(upload>updatarate+0.3)
				estupload = updatarate+0.3;
			else
				estupload = upload;


			// Do not fall under the minUpload allowed
			if(estupload < theApp.glob_prefs->GetMinAutoUpload())
				estupload = theApp.glob_prefs->GetMinAutoUpload();

			// Update Statusline
			char buffer[50];

			if(!nextview){
				if((float)SentStats[m_nArrayIndex].ave/1024>4000000)
					theApp.emuledlg->statusbar.SetText("AutoUpload:no internet connection found",0,0);
					//theApp.emuledlg->AddLogLine(true,"AutoUpload:no internet connection found");
				else if((float)theApp.glob_prefs->GetMaxDownload()>65000)
					//theApp.emuledlg->AddLogLine(true,"AutoUpload: Up: %.1f/%.1f(%.1f)/%.1f | Down: %.1f/full/%.1f (cur/max/all) | Other: %.1f | Protocol: %.1f",updatarate,estupload,upload,(float)SentStats[m_nArrayIndex].ave/1024,(float)theApp.downloadqueue->GetDatarate()/1024,(float)RecvStats[m_nArrayIndex].ave/1024,otherUpload, downUpload);
					//theApp.emuledlg->AddLogLine(true,"AutoUpload: %.1f => %.1f",updatarate,estupload);
					{
					sprintf(buffer,"AutoUpload: %.1f => %.1f",updatarate,estupload);
					theApp.emuledlg->statusbar.SetText(buffer,0,0);
					}
				else{
					//theApp.emuledlg->AddLogLine(true,"AutoUpload: %.1f => %.1f",updatarate,estupload);
					sprintf(buffer,"AutoUpload: %.1f => %.1f",updatarate,estupload);
					theApp.emuledlg->statusbar.SetText(buffer,0,0);
					}
				nextview = 5;
            } else
				nextview--;

			//increment the circular buffer index
			if( ++m_nArrayIndex >= MAX_AVE_TIMES  )
				m_nArrayIndex = 0;

		} else {
			if(isEnabled){
				theApp.emuledlg->AddLogLine(true,"AutoUpload Classic:(disabled)");
				isEnabled = false;
				nextview = 1;
			}
		}

		// Wait for 1 Seconds till next round
		::Sleep(1000);


	}
	isAmuc = true;
	AfxEndThread(0,true);
	return 0;
}

void CAmucThread::StartAMUC(){
	if(Init()){
		theApp.emuledlg->AddLogLine(false,"AMUC Version 0.2 initialized - made by herbert.");
		ResumeThread();
	}
}

bool CAmucThread::Init(){
	if(librariesloaded)
		return true;
	CheckNT();

    m_hInst = LoadLibraryEx( "inetmib1.dll", NULL, 0 );
    if( m_hInst == NULL )
        return false;

    m_fpExtensionInit  = (pSnmpExtensionInit)GetProcAddress( m_hInst ,"SnmpExtensionInit" );
    m_fpExtensionQuery = (pSnmpExtensionQuery)GetProcAddress( m_hInst ,"SnmpExtensionQuery" );

    if( !m_fpExtensionInit || !m_fpExtensionQuery )
        return false;

	//check to see if the MemAlloc and MemFree functions are available
	HINSTANCE hInst = LoadLibraryEx( "SNMPAPI", NULL, 0 );
	if( hInst )
	{
		m_fpSnmpUtilMemAlloc = (SUALLOC)GetProcAddress( hInst,"SnmpUtilMemAlloc" );
		m_fpSnmpUtilMemFree  = (SUFREE)GetProcAddress( hInst, "SnmpUtilMemFree" );
		FreeLibrary( hInst );
	}

	//alloc our bindlist
	m_pvarBindList = (SnmpVarBindList*)SnmpUtilMemAlloc( sizeof(SnmpVarBindList) );
    ASSERT( m_pvarBindList != NULL );

    //init
	librariesloaded = m_fpExtensionInit(GetTickCount(),&hPollForTrapEvent,&SupportedView);
	if(!librariesloaded)
		theApp.emuledlg->AddLogLine(false,"Fatal Error on initialising AMUC! AMUC disabled!");
	return librariesloaded;

}

// Returns the number of bytes received and transmitted through all network interfaces
bool CAmucThread::GetSentAndReceivedOctets( DWORD* pReceived, DWORD *pSent )
{
    RFC1157VarBind      varBind[2];
    AsnInteger          errorStatus;
    AsnInteger          errorIndex;
    int ret;

	if( m_bUse_iphlpapi )
		return( IPHelper( pReceived, pSent ) );

    AsnObjectIdentifier MIB_NULL = {0,0};

    static UINT OID_ifInoctets[] = { 1,3,6,1,2,1,2,2,1,10 };
    AsnObjectIdentifier MIB_ifInoctets = { OID_SIZEOF(OID_ifInoctets), OID_ifInoctets };

    static UINT OID_ifOutoctets[] = { 1,3,6,1,2,1,2,2,1,16 };
    AsnObjectIdentifier MIB_ifOutoctets = { OID_SIZEOF(OID_ifOutoctets), OID_ifOutoctets};

	ASSERT( m_pvarBindList != NULL );

    m_pvarBindList->list = varBind;
    m_pvarBindList->len  = 2;
    varBind[0].name = MIB_NULL;
    varBind[1].name = MIB_NULL;

    SnmpUtilOidCpy(&varBind[0].name,&MIB_ifInoctets);
    SnmpUtilOidCpy(&varBind[1].name,&MIB_ifOutoctets);

    *pReceived = 0;
	*pSent = 0;

    while( TRUE )
    {
		ret = m_fpExtensionQuery( ASN_RFC1157_GETNEXTREQUEST, m_pvarBindList, &errorStatus, &errorIndex );
	    if( !ret )
            break;

		ret = SnmpUtilOidNCmp( &varBind[0].name, &MIB_ifInoctets, MIB_ifInoctets.idLength );
        if( ret != 0 )
            break;

        *pReceived += varBind[0].value.asnValue.number;
        *pSent += varBind[1].value.asnValue.number;
    }

    SnmpUtilOidFree( &varBind[0].name );
    SnmpUtilOidFree( &varBind[1].name );

    return 0;
}

void CAmucThread::CalcAverages( float dbTotal, DWORD dwTime, DWORD dwBps, STATS_STRUCT* pStats )
{
    ASSERT( g_nAveragingWindow <= MAX_AVE_TIMES );

    //set current bps
    pStats[m_nArrayIndex].Bps = dwBps;

    //set total bytes sent or recv
    pStats[m_nArrayIndex].total = dbTotal;

    //set the current time (milliseconds)
    pStats[m_nArrayIndex].time = dwTime;

    //start is the index in the array for calculating averages
    int start = m_nArrayIndex - g_nAveragingWindow;
	if( start < 0 )
		start = MAX_AVE_TIMES + start;

    //the array entry may not have been filled in yet
    if( pStats[start].total == 0 )
        start = 0;

    //set average based upon sampling window size
	float dbSampleTotal = 0;

    //total bytes received/sent in our sampling window
    if( pStats[start].total )
        dbSampleTotal = dbTotal - pStats[start].total;

    //elapsed milliseconds
    DWORD dwElapsed = ( pStats[m_nArrayIndex].time - pStats[start].time );

    //calc average
    if( dwElapsed )
        pStats[m_nArrayIndex].ave = MulDiv( (DWORD)dbSampleTotal, 1000, dwElapsed );
}

///////////////////////////////////////////////////////////////////////////////////////////
// Use the IPHLPAPI interface to retrieve the transmitted and received bytes
int CAmucThread::IPHelper( DWORD* pReceived, DWORD *pSent)
{
	int i;
	MIB_IFROW mib;
	ZeroMemory( &mib, sizeof(mib) );

	*pReceived = 0;
	*pSent = 0;

    int interfaces = m_InterfaceArray.GetSize( );
	for( i = 0; i < interfaces; i++ )
	{
		mib.dwIndex = m_InterfaceArray.GetAt( i );
		if( m_fpGetIfEntry( &mib ) == NO_ERROR )
		{
			if(  mib.dwType == 23 ) //only Dial-Up Connector Traffic is considered
			{//theApp.emuledlg->AddLogLine(false,"%s %i",mib.wszName,mib.dwType);

			*pReceived += mib.dwInOctets;
			*pSent += mib.dwOutOctets;
			}
		}
	}

	return( TRUE );
}


LPVOID CAmucThread::SnmpUtilMemAlloc( UINT nSize )
{
	if( m_fpSnmpUtilMemAlloc )
		return( m_fpSnmpUtilMemAlloc( nSize ) );
	else
		return( GlobalAlloc( GMEM_ZEROINIT, nSize ) );
}

void CAmucThread::SnmpUtilMemFree( LPVOID pMem )
{
	if( m_fpSnmpUtilMemFree )
		m_fpSnmpUtilMemFree( pMem );
	else
		GlobalFree( pMem );
}

DWORD CAmucThread::GetServicePack( )
{
	CRegKey key;
	#define SZ_SPKEY "System\\CurrentControlSet\\Control\\Windows"

    DWORD dwVersion = 0;
    if( key.Open( HKEY_LOCAL_MACHINE, SZ_SPKEY ) == ERROR_SUCCESS )
    {
        key.QueryValue( dwVersion, "CSDVersion" );
		key.Close();
    }

    return( dwVersion );
}

//if running under NT use iphlpapi -- requires SP4 for NT4 otherwise there is a memory leak in SNMP
//We could use this on Win98, however certain releases of IE5 cause iphlpapi to fail.
bool CAmucThread::CheckNT( )
{

	//check for NT
    DWORD dwInterfaces = 0;
	OSVERSIONINFO os;
	os.dwOSVersionInfoSize = sizeof( os );
	GetVersionEx( &os );

	if( os.dwPlatformId == VER_PLATFORM_WIN32_NT )
	{
		m_hInstIpHlp = LoadLibraryEx( "iphlpapi.dll", NULL, 0 );
		if( m_hInstIpHlp )
		{
		    m_fpGetIfEntry = (fpGetIfEntry)GetProcAddress(m_hInstIpHlp,"GetIfEntry" );

            if( os.dwMajorVersion < 5 )
            {
                //requires SP4 or higher
                DWORD dwVersion = GetServicePack( );
                if( dwVersion > 0x300 )
                {
                    //WinNT 4 uses scalar values for the GetIfEntry function
                    //although none of this seems to be documented yet by Microsoft
			        fpGetNumberOfInterfaces fp = (fpGetNumberOfInterfaces)GetProcAddress( m_hInstIpHlp, "GetNumberOfInterfaces" );
			        if( fp )
                    {
                        if( fp( &dwInterfaces ) == NO_ERROR )
                        {
                            m_InterfaceArray.SetSize( dwInterfaces );
                            for( DWORD x = 0; x < dwInterfaces; x++ )
                                m_InterfaceArray.SetAt( x, x+1 ); //not zero based
                        }
                    }
                }

            } else {
                //windows 2000 and win98 use what appear to be scalar values in the lower word
                //and control flags in the upper word of the adapter[].index
                DWORD dwSize = 0;
                PIP_INTERFACE_INFO pInterface;
                typedef DWORD (WINAPI* fpGetInterfaceInfo)( PIP_INTERFACE_INFO, LPDWORD );
                fpGetInterfaceInfo fpInterfaces = (fpGetInterfaceInfo)GetProcAddress( m_hInstIpHlp, "GetInterfaceInfo" );

                if( fpInterfaces )
                {
                    fpInterfaces( NULL, &dwSize );

                    if( dwSize )
                    {
                        pInterface = (PIP_INTERFACE_INFO)GlobalAlloc( GPTR, dwSize );
                        fpInterfaces( pInterface, &dwSize );

                        m_InterfaceArray.SetSize( pInterface->NumAdapters );
                        dwInterfaces = pInterface->NumAdapters;
                        for( DWORD i = 0; i < dwInterfaces; i++ )
                        {
                            m_InterfaceArray.SetAt( i, pInterface->Adapter[i].Index );
                        }
                        GlobalFree( pInterface );
                    }
                }
            }
		}
	}

	m_bUse_iphlpapi = ( dwInterfaces > 0 );
	return( m_bUse_iphlpapi );
}
