//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#pragma once
#include <windows.h>
#include <winsock.h>
#include "types.h"
#include <snmp.h>

#define OID_SIZEOF( Oid )( sizeof Oid / sizeof(UINT) )
#define MAX_AVE_TIMES 61

typedef struct
{
    DWORD Bps;          //bytes per second          
    DWORD ave;          //average throughput
    DWORD time;         //time in milliseconds that the sample was taken
    float total;       //total received or sent in bits
} STATS_STRUCT;

class CAmucThread : public CWinThread
{

typedef BOOL (WINAPI *pSnmpExtensionInit)(
    IN  DWORD               dwTimeZeroReference,
    OUT HANDLE              *hPollForTrapEvent,
    OUT AsnObjectIdentifier *supportedView);
	
typedef BOOL (WINAPI *pSnmpExtensionTrap)(
    OUT AsnObjectIdentifier *enterprise,
    OUT AsnInteger          *genericTrap,
    OUT AsnInteger          *specificTrap,
    OUT AsnTimeticks        *timeStamp,
    OUT RFC1157VarBindList  *variableBindings);

typedef BOOL (WINAPI *pSnmpExtensionQuery)(
    IN BYTE                   requestType,
    IN OUT RFC1157VarBindList *variableBindings,
    OUT AsnInteger            *errorStatus,
    OUT AsnInteger            *errorIndex);

typedef BOOL (WINAPI *pSnmpExtensionInitEx)(
    OUT AsnObjectIdentifier *supportedView);

typedef LPVOID (CALLBACK* SUALLOC)(UINT);
typedef VOID (CALLBACK* SUFREE)(LPVOID);

typedef DWORD (WINAPI* fpGetNumberOfInterfaces)( LPDWORD );	
typedef DWORD (WINAPI* fpGetIfEntry)( LPVOID );		

DECLARE_DYNCREATE(CAmucThread)

protected:
	CAmucThread();
public:
	~CAmucThread(void);
	float	GetEstimatedUpload(int16 ready = 0)		{clientsready = ready; return estupload*1024;}
	void	StartAMUC();
	virtual int Run(); 
	bool    IsEnabled()					{return isEnabled;}
	virtual	BOOL	InitInstance()		{return true;}

private:
	bool Init();
    void CalcAverages( float dwTotal, DWORD dwTime, DWORD dwBPS, STATS_STRUCT* pStats );
	bool GetSentAndReceivedOctets(DWORD* pReceived, DWORD *pSent);
	int  IPHelper(DWORD* pReceived, DWORD *pSent);
	bool CheckNT();
	
	//overloaded functions to accommodate Win95
	LPVOID SnmpUtilMemAlloc( UINT nSize );
	void  SnmpUtilMemFree( LPVOID pMem );
	
	STATS_STRUCT RecvStats[MAX_AVE_TIMES];
    STATS_STRUCT SentStats[MAX_AVE_TIMES];

	float estupload;
	int g_nAveragingWindow;
	int m_nArrayIndex;
	DWORD m_dwStartTime;
    float m_dbTotalBytesRecv;
    float m_dbTotalBytesSent;
	
    HINSTANCE             m_hInst,m_hInstIpHlp;
	pSnmpExtensionInit    m_fpExtensionInit;
	pSnmpExtensionQuery   m_fpExtensionQuery;
	fpGetIfEntry		  m_fpGetIfEntry;	

	SUALLOC m_fpSnmpUtilMemAlloc; 
	SUFREE  m_fpSnmpUtilMemFree; 
	RFC1157VarBindList*		m_pvarBindList; 
	BOOL  m_bUse_iphlpapi;
    CUIntArray m_InterfaceArray;
	DWORD GetServicePack( );
	
	int16		clientsready;	// how much clients are ready to put more upload?
	bool		isEnabled;		// is AMUC enabled
	bool		isAmuc;			// is AMUC Thread running ?
	bool		librariesloaded;	// are the libraries loaded
};
