//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "clientcredits.h"
#include "opcodes.h"
#include "emule.h"
#include "safefile.h"
#include <math.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#define CLIENTS_MET_FILENAME	_T("config\\clients.met")

CClientCredits::CClientCredits(CreditStruct* in_credits)
{
	m_pCredits = in_credits;
}

CClientCredits::CClientCredits(uchar* key)
{
	m_pCredits = new CreditStruct;
	memset(m_pCredits, 0, sizeof(CreditStruct));
	md4cpy(m_pCredits->abyKey, key);
}

CClientCredits::~CClientCredits()
{
	delete m_pCredits;
}

void CClientCredits::AddDownloaded(uint32 bytes) {

	//encode
	uint64 current=m_pCredits->nDownloadedHi;
	current=(current<<32)+ m_pCredits->nDownloadedLo + bytes ;

	//recode
	m_pCredits->nDownloadedLo=(uint32)current;
	m_pCredits->nDownloadedHi=(uint32)(current>>32);
}

void CClientCredits::AddUploaded(uint32 bytes) {

	//encode
	uint64 current=m_pCredits->nUploadedHi;
	current=(current<<32)+ m_pCredits->nUploadedLo + bytes ;

	//recode
	m_pCredits->nUploadedLo=(uint32)current;
	m_pCredits->nUploadedHi=(uint32)(current>>32);
}

uint64	CClientCredits::GetUploadedTotal(){
	return ( (uint64)m_pCredits->nUploadedHi<<32)+m_pCredits->nUploadedLo;
}

uint64	CClientCredits::GetDownloadedTotal(){
	return ( (uint64)m_pCredits->nDownloadedHi<<32)+m_pCredits->nDownloadedLo;
}

float CClientCredits::GetScoreRatio()
{
  // new creditsystem by [lovelace]
  double cl_up,cl_down;
  float result;
  if (this->m_pCredits)
  {
    cl_up=GetUploadedTotal()/(double)1048576;
    cl_down=GetDownloadedTotal()/(double)1048576;
    result=(float)(3.0*cl_down*cl_down-cl_up*cl_up);
    if (fabs(result)>10000.0f) result*=10000.0f/fabs(result);
    result=100.0f*powf((float)(1-1/(1.0f+expf(result*0.001))),6.6667f);
    if (result<0.1f) result=0.1f;
    return (result);
  }
  return 0.1f;
  // end new creditsystem by [lovelace]
}

CClientCreditsList::CClientCreditsList(CPreferences* in_prefs)
{
	m_pAppPrefs = in_prefs;
	m_nLastSaved = ::GetTickCount();
	LoadList();
}

CClientCreditsList::~CClientCreditsList()
{
	SaveList();
	CClientCredits* cur_credit;
	CCKey tmpkey(0);
	POSITION pos = m_mapClients.GetStartPosition();
	while (pos){
		m_mapClients.GetNextAssoc(pos, tmpkey, cur_credit);
		delete cur_credit;
	}
	m_mapClients.RemoveAll();
}

void CClientCreditsList::LoadList()//>>modified by lovelace
{
	CSafeFile file, file2;
  try{
		CString strFileName = m_pAppPrefs->GetAppDir() + CString("clients.met");
		if (!file.Open(strFileName, CFile::modeRead)){
			theApp.emuledlg->AddLogLine(true, GetResString(IDS_ERR_LOADCREDITFILE));
			return;
		}
		CString strFile2Name = m_pAppPrefs->GetAppDir() + CString("trusted.met");
		if (!file2.Open(strFile2Name, CFile::modeRead))
    {
      theApp.emuledlg->AddLogLine(true, "The credit file will get expanded by trusted.met.");
      if (!file2.Open(strFile2Name, CFile::modeWrite|CFile::modeCreate))
      {
        theApp.emuledlg->AddLogLine(true, GetResString(IDS_ERR_FAILED_CREDITSAVE));
        return;
      }
		}
		uint8 version;
		file.Read(&version, 1);
		if (version != CREDITFILE_VERSION){
			theApp.emuledlg->AddLogLine(false, GetResString(IDS_ERR_CREDITFILEOLD));
			file.Close();
			return;
		}

		// everything is ok, lets see if the backup exist...
		CString strBakFileName;
		strBakFileName.Format("%sclients.met.BAK", m_pAppPrefs->GetAppDir());
		
		DWORD dwBakFileSize = 0;
		BOOL bCreateBackup = TRUE;

		HANDLE hBakFile = ::CreateFile(strBakFileName, GENERIC_READ, FILE_SHARE_READ, NULL,
										OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

		if (hBakFile != INVALID_HANDLE_VALUE)
		{
			// Ok, the backup exist, get the size
			dwBakFileSize = ::GetFileSize(hBakFile, NULL); //debug
			if (dwBakFileSize > (DWORD)file.GetLength())
			{
				// the size of the backup was larger then the org. file, something is wrong here, don't overwrite old backup..
				bCreateBackup = FALSE;
			}
			//else: backup is smaller or the same size as org. file, proceed with copying of file
			::CloseHandle(hBakFile);
		}
		//else: the backup doesn't exist, create it

		if (bCreateBackup)
		{
			file.Close(); // close the file before copying
			BOOL bResult = ::CopyFile(strFileName, strBakFileName, FALSE);
			DWORD dwError;
			if (!bResult)
			{
				dwError = ::GetLastError(); // debug
				theApp.emuledlg->AddLogLine(false, GetResString(IDS_ERR_MAKEBAKCREDITFILE));
			}
			// reopen file
			if (!file.Open(strFileName, CFile::modeRead))
			{
				theApp.emuledlg->AddLogLine(true, GetResString(IDS_ERR_LOADCREDITFILE));
				return;
			}
			file.Seek(1, CFile::begin); //set filepointer
		}


		uint32 count;
		CreditFileStruct* cfstruct=new CreditFileStruct;
    file.Read(&count, 4);
		m_mapClients.InitHashTable(count+5000); // TODO: should be prime number... and 20% larger

		for (uint32 i = 0; i < count; i++){
			CreditStruct* newcstruct = new CreditStruct;
			memset(newcstruct, 0, sizeof(CreditStruct));
			memset(cfstruct, 0, sizeof(CreditFileStruct));
			file.Read(cfstruct, sizeof(CreditFileStruct));
	    memcpy(newcstruct->abyKey, cfstruct->abyKey, 16);
      newcstruct->nUploadedLo=cfstruct->nUploadedLo;
      newcstruct->nDownloadedLo=cfstruct->nDownloadedLo; 
      newcstruct->nLastSeen=cfstruct->nLastSeen;
      newcstruct->nUploadedHi=cfstruct->nUploadedHi;
      newcstruct->nDownloadedHi=cfstruct->nDownloadedHi; 
      newcstruct->nReserved3=cfstruct->nReserved3;
      if (file2.GetPosition()+sizeof(uint32)<=file2.GetLength())
        file2.Read(&newcstruct->trustedIP, sizeof(uint32));
      else // maybe not needed?
        newcstruct->trustedIP=0; // maybe not needed?
      CClientCredits* newcredits = new CClientCredits(newcstruct);
			m_mapClients.SetAt(CCKey(newcredits->GetKey()), newcredits);
		}
		delete [] cfstruct;
    file.Close();
		theApp.emuledlg->AddLogLine(false, GetResString(IDS_CREDITFILELOADED), count);
    if (!theApp.stat_clientcredits) theApp.stat_clientcredits=count;//lovelace
	}
	catch(CFileException* error){
		OUTPUT_DEBUG_TRACE();
		if (error->m_cause == CFileException::endOfFile)
			theApp.emuledlg->AddLogLine(true, GetResString(IDS_CREDITFILECORRUPT));
		else{
			char buffer[150];
			error->GetErrorMessage(buffer, 150);
			theApp.emuledlg->AddLogLine(true, GetResString(IDS_ERR_CREDITFILEREAD), buffer);
		}
		error->Delete();
  }
}//<<modified by lovelace

void CClientCreditsList::SaveList()//>>modified by lovelace
{
	uint32 count = m_mapClients.GetCount();
	BYTE* pBuffer = new BYTE[count*sizeof(CreditFileStruct)];
	BYTE* pBuffer2 = new BYTE[count*sizeof(uint32)];
	CString name = m_pAppPrefs->GetAppDir() + CString("clients.met");
	CString name2 = m_pAppPrefs->GetAppDir() + CString("trusted.met");
	CFile file, file2;

	if (!file.Open(name, CFile::modeWrite|CFile::modeCreate))
	{
		theApp.emuledlg->AddLogLine(true, GetResString(IDS_ERR_FAILED_CREDITSAVE));
		return;
	}
	if (!file2.Open(name2, CFile::modeWrite|CFile::modeCreate))
	{
		theApp.emuledlg->AddLogLine(true, GetResString(IDS_ERR_FAILED_CREDITSAVE));
		return;
	}

	CClientCredits* cur_credit;
  CreditFileStruct*	filecredit=new CreditFileStruct; //lovelace
  uchar null_hash[16];
  memset(null_hash, 0, sizeof(null_hash));
  CCKey tempkey(0);
  POSITION pos = m_mapClients.GetStartPosition();
  count = 0;
	while (pos)
	{
		m_mapClients.GetNextAssoc(pos, tempkey, cur_credit);
	    memcpy(filecredit->abyKey, cur_credit->GetDataStruct()->abyKey, 16);
		if ((cur_credit->GetUploadedTotal() || cur_credit->GetDownloadedTotal())
      && (memcmp(null_hash,filecredit->abyKey,sizeof(filecredit->abyKey))!=0))
    {
        filecredit->nUploadedLo=cur_credit->GetDataStruct()->nUploadedLo;
        filecredit->nDownloadedLo=cur_credit->GetDataStruct()->nDownloadedLo; 
        filecredit->nLastSeen=cur_credit->GetDataStruct()->nLastSeen;
        filecredit->nUploadedHi=cur_credit->GetDataStruct()->nUploadedHi;
        filecredit->nDownloadedHi=cur_credit->GetDataStruct()->nDownloadedHi; 
        filecredit->nReserved3=cur_credit->GetDataStruct()->nReserved3;
        memcpy(pBuffer+(count*sizeof(CreditFileStruct)), filecredit, sizeof(CreditFileStruct));
        memcpy(pBuffer2+(count*sizeof(uint32)),&(cur_credit->GetDataStruct()->trustedIP),sizeof(uint32));
        count++; 
		}
	}
  delete [] filecredit;

	uint8 version = CREDITFILE_VERSION;
	file.Write(&version, 1);
	file.Write(&count, 4);
	file.Write(pBuffer, count*sizeof(CreditFileStruct));
	file2.Write(pBuffer2, count*sizeof(uint32));
	file.Close();
	file2.Close();

	delete[] pBuffer;
	delete[] pBuffer2;
	m_nLastSaved = ::GetTickCount();
  theApp.stat_clientcredits=count;
}//<<modified by lovelace

CClientCredits* CClientCreditsList::GetCredit(uchar* key)
{
	CClientCredits* result;
	CCKey tkey(key);
	if (!m_mapClients.Lookup(tkey, result)){
		result = new CClientCredits(key);
		m_mapClients.SetAt(CCKey(result->GetKey()), result);
	}
	result->SetLastSeen();
	return result;
}

void CClientCreditsList::Process()
{
	if (::GetTickCount() - m_nLastSaved > 70000)
		SaveList();
}
