// Written by enkeyDEV(Ottavio84)

#include "StdAfx.h"
#include "emule.h"
#include "releasebar.h"
#include "knownfile.h"

CReleaseBar::CReleaseBar(CKnownFile* file)
{
	m_pFile = file;
	dwLastUpdated = 0;
	m_modified = true;
	m_bISpread = false;
	m_arChunkAssignedTo = 0;
}

CReleaseBar::~CReleaseBar(void)
{
	if (m_arChunkAssignedTo)
		delete [] m_arChunkAssignedTo;
}

template<class T>
__inline void Swap(T &value1, T &value2)
{
	T tmp = value1;
	value1 = value2;
	value2 = tmp;
}

#define ITEM(x) blocklist.GetAt(x)

void CReleaseBar::InsertBlock(CReleaseBar::SentBlock newblock)
{
	m_modified = true;
	// If an empty block, don't insert it
	if (newblock.start == newblock.end)
		return;
	// Insert the new block on the left of the first block in the list
	// beginning after it.
	POSITION pos = blocklist.GetHeadPosition();
	while (pos) {
		if (ITEM(pos).start > newblock.start)
			break;
		blocklist.GetNext(pos);
	}
	if (pos)
		pos = blocklist.InsertBefore(pos, newblock);
	else
		pos = blocklist.AddTail(newblock);

	// To do the smallest possible number of iterations, we start from the block 
	// left to the inserted one, and stop the cycle when there are no more 
	// overlapping blocks
	blocklist.GetPrev(pos);
	if (!pos)
		pos = blocklist.GetHeadPosition();
	while (1) {
		POSITION posnext = pos;
		blocklist.GetNext(posnext);
		if (!posnext)
			break;
		// Look for ovelapping blocks. If there are any, we split them and assign the right frequency.
		if (ITEM(posnext).start < ITEM(pos).end) {
			Swap(ITEM(posnext).start, ITEM(pos).end);
			unsigned int newfreq = ITEM(pos).frequency + ITEM(posnext).frequency;
			if (newfreq > 11)	// The status bar becomes deep blue with a frequency over 10, so it's unuseful to go 
				newfreq = 11;	// any further. This way, more blocks will be merged and so a bit of space is saved.
			// Current block completely overlapped the next, so we adjust it and assign the frequency of the current.
			if (ITEM(posnext).start > ITEM(posnext).end) {
				Swap(ITEM(posnext).start, ITEM(posnext).end);
				ITEM(posnext).frequency = ITEM(pos).frequency;
			}
			SentBlock nwblock = {ITEM(pos).end, 
				ITEM(posnext).start, 
				 newfreq};
			POSITION newpos = blocklist.InsertAfter(pos, nwblock);
			// Remove empty blocks (they are created by the previous operation in most cases)
			if (ITEM(pos).start == ITEM(pos).end) {
				blocklist.RemoveAt(pos);
				pos = newpos;
			}
			if (ITEM(posnext).start == ITEM(posnext).end) {
				blocklist.RemoveAt(posnext);
				posnext = newpos;
			}
		} else
            // No more overlapping blocks, we can exit the loop
			break;

		pos = posnext;
	}
	
	// Merge adjacent blocks with equal frequency: the list must be saved, so it must be 
	// as short as possible.
	pos = blocklist.GetHeadPosition();
	while(1) {
        POSITION posnext = pos;
		blocklist.GetNext(posnext);
		if (!posnext)
			break;
		if ((ITEM(pos).end == ITEM(posnext).start) && (ITEM(pos).frequency == ITEM(posnext).frequency)) {
			ITEM(posnext).start = ITEM(pos).start;
			blocklist.RemoveAt(pos);
		}
		pos = posnext;
	}
}

void CReleaseBar::DrawBar(CDC* dc, RECT *rect, bool bFlat)
{
	COLORREF crProgress; 
	COLORREF crHave; 
	COLORREF crPending;
	COLORREF crMissing = RGB(255, 0, 0); 
	unsigned int completed = 0;

	if(bFlat) { 
		crProgress = RGB(0, 150, 0); 
		crHave = RGB(0, 0, 0); 
		crPending = RGB(255,255,100);
	} else { 
		crProgress = RGB(0, 224, 0); 
		crHave = RGB(104, 104, 104); 
		crPending = RGB(255, 208, 0);
	} 

	s_ChunkBar.SetHeight(rect->bottom - rect->top); 
	s_ChunkBar.SetWidth(rect->right - rect->left); 
	s_ChunkBar.SetFileSize(m_pFile->GetFileSize()); 
	s_ChunkBar.Fill(crMissing); 

	for (POSITION pos = blocklist.GetHeadPosition(); pos; blocklist.GetNext(pos)) {
		COLORREF color;
		int freq = ITEM(pos).frequency;
		color = RGB(0,(210-(22*(freq-1)) <  0)? 0:210-(22*(freq-1)),255); 
		s_ChunkBar.FillRange(ITEM(pos).start, ITEM(pos).end, color);
		completed += ITEM(pos).end - ITEM(pos).start;
	}

	s_ChunkBar.Draw(dc, rect->left, rect->top, bFlat); 
	// green progress 
	float blockpixel = (float)(rect->right -  rect->left)/((float)m_pFile->GetFileSize()); 
	RECT gaprect; 
	gaprect.top = rect->top; 
	gaprect.bottom = gaprect.top + 3; 
	gaprect.left = rect->left; 

	if(!bFlat) { 
		s_LoadBar.SetWidth((uint32)((float)(completed)*blockpixel + .5f)); 
		s_LoadBar.SetHeight(4);
		s_LoadBar.Fill(crProgress); 
		s_LoadBar.Draw(dc, gaprect.left, gaprect.top, false); 
	} else { 
		gaprect.right = rect->left + (uint32)((float)(completed)*blockpixel + .5f); 
		dc->FillRect(&gaprect, &CBrush(crProgress)); 
		//draw gray progress only if flat 
		gaprect.left = gaprect.right; 
		gaprect.right = rect->right; 
		dc->FillRect(&gaprect, &CBrush(RGB(224,224,224)));
	} 
}

#define FT_REL_FREQ			0x60
#define FT_REL_START		0x61
#define	FT_REL_END			0x62
#define	FT_REL_ACTIVATED	0x63
#define FT_ISPREAD			0x64

bool CReleaseBar::LoadFromTagList(CArray<CTag*,CTag*>* taglist)
{
	CTag *tag1, *tag2, *tag3;
	SentBlock newblock;
	bool hasreleasebar = false;
	// We go backwards to do less memory operations while removing 
	// items from the array and while inserting new blocks in the
	// list
	int i = taglist->GetUpperBound();
	while (i >= 0) {
		tag1 = taglist->GetAt(i--);
		if (tag1->tag.specialtag != FT_REL_END) {//Changed By Yun.SF3
			if (tag1->tag.specialtag == FT_REL_ACTIVATED) { //Changed By Yun.SF3
                hasreleasebar = true;
				taglist->RemoveAt(i+1, 1);
				delete tag1;
			} else if (tag1->tag.specialtag == FT_ISPREAD) {//Changed By Yun.SF3
				EnableISpread(true);
				taglist->RemoveAt(i+1, 1);
				delete tag1;
			}
			continue;
		}
		newblock.end = tag1->tag.intvalue;//Changed By Yun.SF3
		
		if (i < 0)
			break;
		tag2 = taglist->GetAt(i--);
		if (tag2->tag.specialtag != FT_REL_START)
			continue;
		newblock.start = tag2->tag.intvalue;//Changed By Yun.SF3	
		
		if (i < 0)
			break;
		tag3 = taglist->GetAt(i--);
		if (tag3->tag.specialtag != FT_REL_FREQ)
			continue;
		newblock.frequency = tag3->tag.intvalue;//Changed By Yun.SF3	

		delete tag1;
		delete tag2;
		delete tag3;
		taglist->RemoveAt(i+1, 3);
		InsertBlock(newblock);
	}
#ifdef DEBUG
	if (!blocklist.IsEmpty())
		theApp.emuledlg->AddLogLine(FALSE, "DEBUG: Loaded %i blocks for the file %s", blocklist.GetCount(), m_pFile->GetFileName());
#endif
	return hasreleasebar;
}
void CReleaseBar::SaveToFile(CFile* file)
{
	CTag* newtag;

	for (POSITION pos = blocklist.GetHeadPosition(); pos; blocklist.GetNext(pos)) {
		SentBlock curblock = ITEM(pos);
		newtag = new CTag(FT_REL_FREQ, curblock.frequency);
		newtag->WriteTagToFile(file);
		delete newtag;
		newtag = new CTag(FT_REL_START, curblock.start);
		newtag->WriteTagToFile(file);
		delete newtag;
		newtag = new CTag(FT_REL_END, curblock.end);
		newtag->WriteTagToFile(file);
		delete newtag;
	}
	newtag = new CTag(FT_REL_ACTIVATED, (uint32)0);
	newtag->WriteTagToFile(file);
	delete newtag;
	if (m_bISpread) {
		newtag = new CTag(FT_ISPREAD, (uint32)0);
		newtag->WriteTagToFile(file);
		delete newtag;
	}
#ifdef DEBUG
	if (!blocklist.IsEmpty())
        theApp.emuledlg->AddLogLine(FALSE, "Saved %i blocks for the file %s", blocklist.GetCount(), m_pFile->GetFileName());
#endif

}

void CReleaseBar::SentData(Requested_Block_Struct* sentdata) 
{
	SentBlock sentblock = {sentdata->StartOffset, sentdata->EndOffset+1, 1}; 
	InsertBlock(sentblock);	
	theApp.sharedfiles->UpdateFile(m_pFile);
}
float CReleaseBar::CalculateDoneRatio(void)
{
	uint32 completed = 0;
	uint32 totalsent = 0;
	if (m_modified) {
		m_modified = false;
		if (!m_pFile->GetFileSize())
			m_doneratio = 1;
		else {
			for (POSITION pos = blocklist.GetHeadPosition(); pos; blocklist.GetNext(pos)) {
				completed += ITEM(pos).end - ITEM(pos).start;
				totalsent += (ITEM(pos).end - ITEM(pos).start) * ITEM(pos).frequency;
			}
			if (completed < m_pFile->GetFileSize())
				m_doneratio = (float)completed / (float)m_pFile->GetFileSize();
			else 
				m_doneratio = (float)totalsent / (float)m_pFile->GetFileSize();
		}
	}
	return m_doneratio;
}

bool CReleaseBar::IsChunkComplete(uint32 chunk)
{
	uint32 start = chunk*PARTSIZE;
	uint32 end = ((chunk+1)*PARTSIZE)-1;

	if (end > m_pFile->GetFileSize())
		end = m_pFile->GetFileSize();
	
	for (POSITION pos = blocklist.GetHeadPosition(); pos != NULL; blocklist.GetNext(pos)) 
		if (ITEM(pos).start <= start) 
			break;

	if (!pos)
		return false;

	uint32 prevend;
	while (1) {	
		if (ITEM(pos).end >= end)
			return true;
		prevend = ITEM(pos).end;
		blocklist.GetNext(pos);
		if (!pos)
			return false;
		if (ITEM(pos).start > prevend)
			return false;
	}
}

void CReleaseBar::EnableISpread(bool bEnable)
{
	if (m_bISpread == bEnable)
		return;
	ASSERT(m_pFile != NULL);
	if (!m_pFile)
		return;
	if (bEnable) {
		m_bISpread = true;
		m_partcount = m_pFile->GetPartCount();
		m_arChunkAssignedTo = new CUpDownClient*[m_partcount];
		memset(m_arChunkAssignedTo, NULL, m_partcount*sizeof(CUpDownClient*));
	} else {
		delete [] m_arChunkAssignedTo;
		m_arChunkAssignedTo = 0;
		m_bISpread = false;
	}
}

#define COMPLETEDCHUNK (CUpDownClient*)0xFFFFFFFF

bool CReleaseBar::CreateFakeFileStatus(CFile* packetdata, CUpDownClient* client)
{
	if (!m_bISpread)
		return false;
	
	int i1stAssignedChunk = -1;
	int i2ndAssignedChunk = -1;
	bool b1stChunkPreviouslyAssigned = false;
	int completedchunks = 0;

	for (int i = 0; i < m_partcount; i++) {
		if (m_arChunkAssignedTo[i] == COMPLETEDCHUNK) {
            completedchunks++;
			continue;
		}
		if (IsChunkComplete(i)) {
			m_arChunkAssignedTo[i] = COMPLETEDCHUNK;
			completedchunks++;
			continue;
		}
		if (client->IsUpPartAvailable(i))
			continue;

		CUpDownClient* cur_client = m_arChunkAssignedTo[i];
		if (cur_client == client) {	//this chunk was already assigned to this client
			if (!b1stChunkPreviouslyAssigned) {
				i1stAssignedChunk = i;
				b1stChunkPreviouslyAssigned = true;
				continue;
			} else {
				i2ndAssignedChunk = i;
				break;
			}
		}
        if (cur_client																									//is an assigned chunk
			&& (theApp.uploadqueue->IsOnUploadQueue(cur_client) || theApp.uploadqueue->IsDownloading(cur_client))		//is still a valid client
			&& !memcmp(cur_client->GetUploadFileID(), m_pFile->GetFileHash(), 16))												//is still requesting the same file
			continue;
		
        // if we arrived here, the i-th chunk is free and can be assigned
		if (i1stAssignedChunk == -1) {
			i1stAssignedChunk = i;
			m_arChunkAssignedTo[i] = client;
		} else if (i2ndAssignedChunk == -1) {
			i2ndAssignedChunk = i;
			m_arChunkAssignedTo[i] = client;
		}
	}
	
	if (completedchunks == m_partcount) {
		EnableISpread(false);
		return false;
	}
	
	packetdata->Write(&m_partcount,2);
	uint16 done = 0;
	if ((i1stAssignedChunk != -1) || (i2ndAssignedChunk != -1)) {
		while (done != m_partcount){
			uint8 towrite = 0;
			for (uint32 i = 0;i != 8;i++){
				if ((done  == i1stAssignedChunk) || (done == i2ndAssignedChunk))
					towrite |= (1<<i);
				done++;
				if (done == m_partcount)
					break;
			}
			packetdata->Write(&towrite,1);
		}
	} else {
		while (done != m_partcount){
			uint8 towrite = 0;
			for (uint32 i = 0;i != 8;i++){
				if (!IsChunkComplete(done))
					towrite |= (1<<i);
				done++;
				if (done == m_partcount)
					break;
			}
			packetdata->Write(&towrite,1);
		}
	}
	
	return true;

}