//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "sharedfilelist.h"
#include "knownfilelist.h"
#include "packets.h"
#include <time.h>
//[Bloodymad Scoring.EstWaiting.Send]
#include "otherfunctions.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


CSharedFileList::CSharedFileList(CPreferences* in_prefs,CServerConnect* in_server,CKnownFileList* in_filelist){
	app_prefs = in_prefs;
	server = in_server;
	filelist = in_filelist;
	output = 0;
	m_Files_map.InitHashTable(1024);
	FindSharedFiles();
	//[Bloodymad Scoring.EstWaiting.Send]
	nReservedSlots = 0;
	nLastReservedSlot = ::GetTickCount();
	nTimePerSlot = 0;
	for(int i=0;i<=MAXPRIORITY;i++){
		nPriorityReservedSlots[i] = 0;
		nPriorityLastReservedSlot[i] = nLastReservedSlot;
		dPriorityTimePerSlotMod[i] = 1.0;
	}
	m_lancast = new CLanCast(in_prefs);	// LANCAST (moosetea) - 
}

// LANCAST (moosetea) - Broadcast the next hash 
void CSharedFileList::NextLANBroadcast(){

static POSITION pos = 0;
static CCKey bufKey;

CKnownFile* cur_file;

if (!m_Files_map.IsEmpty()){

 if (!pos)
  pos = m_Files_map.GetStartPosition();

 if (pos)
 {
	m_Files_map.GetNextAssoc(pos,bufKey,cur_file);
	m_lancast->BroadcastHash(cur_file);
 }
}
}

CSharedFileList::~CSharedFileList(){
	while (!waitingforhash_list.IsEmpty()){
		UnknownFile_Struct* nextfile = waitingforhash_list.RemoveHead();
		delete[] nextfile->directory; 
		delete[] nextfile->name;
		delete nextfile;
	}
	// SLUGFILLER: SafeHash
	while (!currentlyhashing_list.IsEmpty()){
		UnknownFile_Struct* nextfile = currentlyhashing_list.RemoveHead();
		delete[] nextfile->directory; 
		delete[] nextfile->name;
		delete nextfile;
	}
	// SLUGFILLER: SafeHash
	}

void CSharedFileList::FindSharedFiles(){
	if (!m_Files_map.IsEmpty()){
		CSingleLock sLock1(&list_mut,true); // list thread safe
		m_Files_map.RemoveAll();
		sLock1.Unlock();
		theApp.downloadqueue->AddPartFilesToShare(); // read partfiles
	}

  // khaos::kmod+ Fix: Shared files loaded multiple times.
	// TODO: Use a different list or array or something other which does not force use finally produce lowercased directory names
  CStringList l_sAdded;
 
  l_sAdded.AddHead( CString( _strlwr( app_prefs->GetIncomingDir() ) ) );
  AddFilesFromDirectory(l_sAdded.GetHead().GetBuffer());
  if (theApp.glob_prefs->GetCatCount()>1) {
    for (int ix=1;ix<theApp.glob_prefs->GetCatCount();ix++)
    {
      CString tempDir( _strlwr(theApp.glob_prefs->GetCatPath(ix)) );
      if( l_sAdded.Find( tempDir ) ==NULL ) {
        l_sAdded.AddHead( tempDir );
        AddFilesFromDirectory( tempDir.GetBuffer() );
      }
    }
		}

	for (POSITION pos = app_prefs->shareddir_list.GetHeadPosition();pos != 0;app_prefs->shareddir_list.GetNext(pos))
  {
    CString tempDir = app_prefs->shareddir_list.GetAt(pos).MakeLower();
    if( l_sAdded.Find( tempDir ) ==NULL ) {
      l_sAdded.AddHead( tempDir );
      AddFilesFromDirectory(tempDir.GetBuffer());
    }
  }
  // khaos::kmod-
	if (waitingforhash_list.IsEmpty()) 
		theApp.emuledlg->AddLogLine(false,GetResString(IDS_SHAREDFOUND),m_Files_map.GetCount());
	else
		theApp.emuledlg->AddLogLine(false,GetResString(IDS_SHAREDFOUNDHASHING), m_Files_map.GetCount(), waitingforhash_list.GetCount());
	HashNextFile();
}

void CSharedFileList::AddFilesFromDirectory(char* directory){
	CFileFind ff;
	
	char* searchpath = new char[strlen(directory)+3];
	sprintf(searchpath,"%s\\*",directory);
	
	bool end = !ff.FindFile(searchpath,0);
	delete[] searchpath;

	if (end)
		return;

	while (!end){
		end = !ff.FindNextFile();
		if (ff.IsDirectory() || ff.IsDots() || ff.IsSystem() || ff.IsTemporary() || ff.GetLength()>=4294967295 )
			continue;
		CTime lwtime;
		if (!ff.GetLastWriteTime(lwtime))
			theApp.emuledlg->AddDebugLogLine(false, "Failed to get file date of %s - %s", ff.GetFileName(), GetErrorMessage(GetLastError()));
		uint32 fdate = mktime(lwtime.GetLocalTm());
		if (fdate == -1)
			theApp.emuledlg->AddDebugLogLine(false, "Failed to convert file date of %s", ff.GetFileName());
		CKnownFile* toadd = filelist->FindKnownFile(ff.GetFileName().GetBuffer(),fdate,(uint32)ff.GetLength());
		if (toadd){
			toadd->SetPath(directory);
			SafeAddKFile(toadd, true);	// SLUGFILLER: mergeKnown - no unmanagad adds
		}
		else{
			//not in knownfilelist - start adding thread to hash file
                UnknownFile_Struct* tohash = new UnknownFile_Struct;
                tohash->directory = nstrdup(directory);
			tohash->name = nstrdup(ff.GetFileName().GetBuffer());
			// SLUGFILLER: SafeHash - don't double hash
			if (IsHashing(tohash) || theApp.downloadqueue->IsTempFile(tohash)){
				delete[] tohash->directory;
				delete[] tohash->name;
				delete tohash;
			} else
                waitingforhash_list.AddTail(tohash);
			// SLUGFILLER: SafeHash
            }
	}
	ff.Close();
}

void CSharedFileList::SafeAddKFile(CKnownFile* toadd, bool bOnlyAdd){
	RemoveFromHashing(toadd);	// SLUGFILLER: SafeHash - hashed ok, remove from list, in case it was on the list
	CSingleLock sLock(&list_mut,true);
	m_Files_map.SetAt(CCKey(toadd->GetFileHash()),toadd);
	sLock.Unlock();
	if (bOnlyAdd)
		return;
	if (output)
		output->ShowFile(toadd);
	// SLUGFILLER: SafeHash - removed call to HashNextFile, to only be called if a previous hash was finished
	// offer new file to server
	if (!server->IsConnected())
		return;
	CMemFile* files = new CMemFile(100);
	uint32 filecount = 1;
	files->Write(&filecount,4);
	CreateOfferedFilePacket(toadd,files);
	Packet* packet = new Packet(files);
	packet->opcode = OP_OFFERFILES;
	delete files;
	theApp.uploadqueue->AddUpDataOverheadServer(packet->size);
	server->SendPacket(packet,true);
	
}

// removes first occurrence of 'toremove' in 'list'
void CSharedFileList::RemoveFile(CKnownFile* toremove){
	output->RemoveFile(toremove);
	m_Files_map.RemoveKey(CCKey(toremove->GetFileHash()));
}

void CSharedFileList::Reload(bool sendtoserver){
	this->FindSharedFiles();
	if (output)
		output->ShowFileList(this);
	if (sendtoserver)
		SendListToServer();
}

void CSharedFileList::SetOutputCtrl(CSharedFilesCtrl* in_ctrl){
	output = in_ctrl;
	output->ShowFileList(this);
	HashNextFile();		// SLUGFILLER: SafeHash - if hashing not yet started, start it now
}

void CSharedFileList::SendListToServer(){
	if (m_Files_map.IsEmpty() || !server->IsConnected())
		return;
	CSafeMemFile files(1024);
	CCKey bufKey;
	CKnownFile* cur_file,cur_file2;
	POSITION pos,pos2;

	// send all files
	if (server->GetCurrentServer()->GetSoftFiles() && (uint32)m_Files_map.GetCount()<=server->GetCurrentServer()->GetSoftFiles() ) {
	uint32 filecount = m_Files_map.GetCount();
		files.Write(&filecount,4);
		for (pos = m_Files_map.GetStartPosition();pos != 0;){
			m_Files_map.GetNextAssoc(pos,bufKey,cur_file);
			CreateOfferedFilePacket(cur_file,&files);
		}
	} else { // send subset
		CTypedPtrList<CPtrList, CKnownFile*> sortedList;
		bool added=false;
		for(pos=m_Files_map.GetStartPosition(); pos!=0;)
		{
			m_Files_map.GetNextAssoc(pos, bufKey, cur_file);
			added=false;
	
			//insertsort into sortedList
			for (pos2 = sortedList.GetHeadPosition();pos2 != 0 && !added;sortedList.GetNext(pos2)){
				if (GetRealPrio(sortedList.GetAt(pos2)->GetUpPriority()) <= GetRealPrio(cur_file->GetUpPriority()) ) {
					sortedList.InsertBefore(pos2,cur_file);
					added=true;
				}
			}
			if (!added)
				sortedList.AddTail(cur_file);
		}
		// add to packet
		uint32 filecount = server->GetCurrentServer()->GetSoftFiles();
		if (filecount == 0) // if we don't know the soft limit, we send all our files
			filecount = sortedList.GetCount();
		files.Write(&filecount,4);
		uint32 count=0;
		for (pos2 = sortedList.GetHeadPosition();pos2 != 0 && count<filecount ;sortedList.GetNext(pos2)){
			count++;
			CreateOfferedFilePacket(sortedList.GetAt(pos2),&files);
		}
		sortedList.RemoveAll();
	}

	Packet* packet = new Packet(&files);
	packet->opcode = OP_OFFERFILES;
	theApp.uploadqueue->AddUpDataOverheadServer(packet->size);
	server->SendPacket(packet,true);
}

CKnownFile* CSharedFileList::GetFileByIndex(int index){
	int count=0;
	CKnownFile* cur_file;
	CCKey bufKey;

	for (POSITION pos = m_Files_map.GetStartPosition();pos != 0;){
		m_Files_map.GetNextAssoc(pos,bufKey,cur_file);
		if (index==count) return cur_file;
		count++;
	}
	return 0;
}

void CSharedFileList::CreateOfferedFilePacket(CKnownFile* cur_file,CMemFile* files){
	files->Write(cur_file->GetFileHash(),16);
	char* buffer = new char[6];
	memset(buffer,0,6);
	files->Write(buffer,6);
	delete[] buffer;
	files->Write(cur_file->GetFileTypePtr(),4);
	CTag* nametag = new CTag(FT_FILENAME,cur_file->GetFileName());
	nametag->WriteTagToFile(files);
	delete nametag;
	CTag* sizetag = new CTag(FT_FILESIZE,cur_file->GetFileSize());
	sizetag->WriteTagToFile(files);
	delete sizetag;
	//TODO add tags for documents mp3 etc
}

// -khaos--+++> New param:  pbytesLargest, pointer to uint64.
//				Various other changes to accomodate our new statistic...
//				Point of this is to find the largest file currently shared.
uint64 CSharedFileList::GetDatasize(uint64 &pbytesLargest) {
	pbytesLargest=0;
	// <-----khaos-
	uint64 fsize;
	fsize=0;

	CCKey bufKey;
	CKnownFile* cur_file;
	for (POSITION pos = m_Files_map.GetStartPosition();pos != 0;){
		m_Files_map.GetNextAssoc(pos,bufKey,cur_file);
		fsize+=cur_file->GetFileSize();
		// -khaos--+++> If this file is bigger than all the others...well duh.
		if (cur_file->GetFileSize() > pbytesLargest) pbytesLargest = cur_file->GetFileSize();
		// <-----khaos-
	}
	return fsize;
}

CKnownFile*	CSharedFileList::GetFileByID(uchar* filehash){
	CKnownFile* result;
	CCKey tkey(filehash);
	if (filehash && m_Files_map.Lookup(tkey,result))
		return result;
	else
		return 0;
}

void CSharedFileList::HashNextFile(){
	// SLUGFILLER: SafeHash
	if (!::IsWindow(theApp.emuledlg->m_hWnd))	// wait for the dialog to open
		return;
	if (!currentlyhashing_list.IsEmpty())	// one hash at a time
		return;
	// SLUGFILLER: SafeHash
	if (waitingforhash_list.IsEmpty())
		return;
	UnknownFile_Struct* nextfile = waitingforhash_list.RemoveHead();
	currentlyhashing_list.AddTail(nextfile);	// SLUGFILLER: SafeHash - keep track
	CAddFileThread* addfilethread = (CAddFileThread*) AfxBeginThread(RUNTIME_CLASS(CAddFileThread), THREAD_PRIORITY_NORMAL,0, CREATE_SUSPENDED);	// SLUGFILLER: SafeHash - full speed hashing
	addfilethread->SetValues(this,nextfile->directory,nextfile->name);
	addfilethread->ResumeThread();
	// SLUGFILLER: SafeHash - nextfile deleting handled elsewhere
	
}

//[Bloodymad Scoring.EstWaiting.Send]
uint32 CSharedFileList::GetTimePerSlot() {
	uint32 actual = ::GetTickCount()-nLastReservedSlot;
	uint32 esttime = nTimePerSlot;
	if (esttime>=actual)
		return esttime;
	else
		return slidingAverage(esttime,actual,boundinvert(nReservedSlots+1,25));
}

uint32 CSharedFileList::GetTimePerSlot(int8 filepriority) {
	uint32 actual = ::GetTickCount()-nPriorityLastReservedSlot[filepriority];
	uint32 esttime = GetTimePerSlot()*dPriorityTimePerSlotMod[filepriority];
	if (esttime>=actual)
		return esttime;
	else
		return slidingAverage(esttime,actual,boundinvert(nPriorityReservedSlots[filepriority]+1,15));
}

void CSharedFileList::RegisterReservedSlot(int8 filepriority, uint32 thisreservedslot, uint32 thiswaittime) {
	// We are not going back in time
	ASSERT(thisreservedslot>=nLastReservedSlot);
	// What if == ?

	// Update general information
	nReservedSlots++;
	uint32 waitedslot = thisreservedslot-nLastReservedSlot;
	// lets assume that our upload slots are always filled and there is a queue.
	// otherwise it's at all pointless to calculate waiting times...
	nTimePerSlot = slidingAverage(
		nTimePerSlot,
		waitedslot,
		boundinvert(nReservedSlots,25));
	nLastReservedSlot = thisreservedslot;

	// Update priority specific information
	nPriorityReservedSlots[filepriority]++;
	uint32 prioritywaitedslot = thisreservedslot-nPriorityLastReservedSlot[filepriority];
	if (thiswaittime<prioritywaitedslot)
		prioritywaitedslot = thiswaittime;
	dPriorityTimePerSlotMod[filepriority] = slidingAverage(
		dPriorityTimePerSlotMod[filepriority],
		(double)prioritywaitedslot/GetTimePerSlot(),
		boundinvert(nPriorityReservedSlots[filepriority],15));
	nPriorityLastReservedSlot[filepriority] = thisreservedslot;
}
	
// SLUGFILLER: SafeHash
bool CSharedFileList::IsHashing(UnknownFile_Struct* tohash){
	for (POSITION pos = waitingforhash_list.GetHeadPosition();pos != 0;waitingforhash_list.GetNext(pos))
		if (!stricmp(waitingforhash_list.GetAt(pos)->name, tohash->name) && !stricmp(waitingforhash_list.GetAt(pos)->directory, tohash->directory))
			return true;
	for (POSITION pos = currentlyhashing_list.GetHeadPosition();pos != 0;currentlyhashing_list.GetNext(pos))
		if (!stricmp(currentlyhashing_list.GetAt(pos)->name, tohash->name) && !stricmp(currentlyhashing_list.GetAt(pos)->directory, tohash->directory))
			return true;
	return false;
}

void CSharedFileList::RemoveFromHashing(CKnownFile* hashed){
	for (POSITION pos = currentlyhashing_list.GetHeadPosition();pos != 0;currentlyhashing_list.GetNext(pos))
		if (!stricmp(currentlyhashing_list.GetAt(pos)->name, hashed->GetFileName()) && !stricmp(currentlyhashing_list.GetAt(pos)->directory, hashed->GetPath())){
			UnknownFile_Struct* removefile = currentlyhashing_list.GetAt(pos);
			currentlyhashing_list.RemoveAt(pos);
			delete[] removefile->directory;
			delete[] removefile->name;
			delete removefile;
			HashNextFile();			// start next hash if possible, but only if a previous hash finished
			return;
		}
}

void CSharedFileList::HashFailed(UnknownFile_Struct* hashed){
	for (POSITION pos = currentlyhashing_list.GetHeadPosition();pos != 0;currentlyhashing_list.GetNext(pos))
		if (!stricmp(currentlyhashing_list.GetAt(pos)->name, hashed->name) && !stricmp(currentlyhashing_list.GetAt(pos)->directory, hashed->directory)){
			UnknownFile_Struct* removefile = currentlyhashing_list.GetAt(pos);
			currentlyhashing_list.RemoveAt(pos);
			delete[] removefile->directory;
			delete[] removefile->name;
			delete removefile;
			HashNextFile();			// start next hash if possible, but only if a previous hash finished
			break;
		}
	delete[] hashed->directory;
	delete[] hashed->name;
	delete hashed;
}
// SLUGFILLER: SafeHash

IMPLEMENT_DYNCREATE(CAddFileThread, CWinThread)
CAddFileThread::CAddFileThread(){
	m_pOwner = 0;
	filename = 0;
	directory = 0;
}
void CAddFileThread::SetValues(CSharedFileList* pOwner, char* in_directory, char* in_filename, CPartFile* in_partfile_Owner){
	 m_pOwner = pOwner;
	 directory = nstrdup(in_directory);
	 filename = nstrdup(in_filename);
	 partfile_Owner = in_partfile_Owner;
}

int CAddFileThread::Run(){
	DbgSetThreadName("Hashing %s", filename);
	if (!(m_pOwner || partfile_Owner) || !filename)
		AfxEndThread(0,true);
	// SLUGFILLER: SafeHash - locking code removed, unnececery
	CKnownFile* newrecord = new CKnownFile();
	if (newrecord->CreateFromFile(directory,filename) && theApp.emuledlg->IsRunning()){	// SLUGFILLER: SafeHash - in case of shutdown while still hashing
		PostMessage(theApp.emuledlg->m_hWnd,TM_FINISHEDHASHING,(m_pOwner ? 0:(WPARAM)partfile_Owner),(LPARAM)newrecord);
	}
	else{
		// SLUGFILLER: SafeHash - inform main program of hash failure
		if (m_pOwner && theApp.emuledlg->IsRunning()){
			UnknownFile_Struct* hashed = new UnknownFile_Struct;
			hashed->directory = nstrdup(directory);
			hashed->name = nstrdup(filename);
			PostMessage(theApp.emuledlg->m_hWnd,TM_HASHFAILED,0,(LPARAM)hashed);
			delete newrecord;
		}
		// SLUGFILLER: SafeHash
		delete newrecord;
	}

	delete[] filename;
	if (directory)
		delete[] directory;
	// SLUGFILLER: SafeHash - locking code removed, unnececery
	AfxEndThread(0,true);
	return 0;
}

void CSharedFileList::UpdateFile(CKnownFile* toupdate) {output->UpdateFile(toupdate);}
