#include "StdAfx.h"
#include "sourcesaver.h"
#include "PartFile.h"
#include "emule.h"


#define RELOADTIME	3600000 //60 minutes	
#define RESAVETIME	 600000 //10 minutes


CSourceSaver::CSourceSaver(void)
{
m_dwLastTimeLoaded = ::GetTickCount() - RELOADTIME;
m_dwLastTimeSaved = ::GetTickCount() + (rand() * 30000 / RAND_MAX) - 15000 - RESAVETIME;

}

CSourceSaver::CSourceData::CSourceData(CUpDownClient* client, const char* exp) 
{
	sourceID = client->GetUserID(); 
	sourcePort = client->GetUserPort(); 
	partsavailable = client->GetAvailablePartCount();
	memcpy(expiration, exp, 7);
	expiration[6] = 0;
}


CSourceSaver::~CSourceSaver(void)
{
}

bool CSourceSaver::Process(CPartFile* file, int maxSourcesToSave) // return false if sources not saved
{
	if ((int)(::GetTickCount() - m_dwLastTimeSaved) > RESAVETIME) {
		m_dwLastTimeSaved = ::GetTickCount() + (rand() * 30000 / RAND_MAX) - 15000;
		SourceList srcs;
		CString slsfilepath;
		slsfilepath.Format("%s\\%s.txtsrc", theApp.glob_prefs->GetTempDir(), file->GetPartMetFileName());
		LoadSourcesFromFile(file, &srcs, slsfilepath);
		SaveSources(file, &srcs, slsfilepath, maxSourcesToSave);
		
		if ((int)(::GetTickCount() - m_dwLastTimeLoaded) > RELOADTIME) {
			m_dwLastTimeLoaded = ::GetTickCount() + (rand() * 30000 / RAND_MAX) - 15000;
			AddSourcesToDownload(file, &srcs);
		}

		while (!srcs.IsEmpty()) 
			delete srcs.RemoveHead();
		
		return true;
	}
	return false;
}

void CSourceSaver::DeleteFile(CPartFile* file)
{
	CString slsfilepath;
	slsfilepath.Format("%s\\%s.txtsrc", theApp.glob_prefs->GetTempDir(), file->GetPartMetFileName());
	if (remove(slsfilepath)) if (errno != ENOENT)
		theApp.emuledlg->AddLogLine(true, "Failed to delete %s, you will need to do this by hand", slsfilepath);    
}

void CSourceSaver::LoadSourcesFromFile(CPartFile* file, SourceList* sources, CString& slsfile)
{
	CString strLine;
	CStdioFile f;
	if (!f.Open(slsfile, CFile::modeRead | CFile::typeText))
		return;
	while(f.ReadString(strLine)) {
		if (strLine.GetAt(0) == '#')
			continue;
		int pos = strLine.Find(':');
		if (pos == -1)
			continue;
		CString strIP = strLine.Left(pos);
		strLine = strLine.Mid(pos+1);
		uint32 dwID = inet_addr(strIP);
		if (dwID == INADDR_NONE) 
			continue;
		pos = strLine.Find(',');
		if (pos == -1)
			continue;
		CString strPort = strLine.Left(pos);
		strLine = strLine.Mid(pos+1);
		uint16 wPort = atoi(strPort);
		if (!wPort)
			continue;
		pos = strLine.Find(';');
		if (pos == -1)
			continue;
		CString strExpiration = strLine.Left(pos);
		if (IsExpired(strExpiration))
			continue;

		CSourceData* newsource = new CSourceData(dwID, wPort, strExpiration);
		sources->AddTail(newsource);
		
	}
    f.Close();
}

void CSourceSaver::AddSourcesToDownload(CPartFile* file, SourceList* sources) 
{
	for (POSITION pos = sources->GetHeadPosition(); pos; sources->GetNext(pos)) {
		if (theApp.glob_prefs->GetMaxSourcePerFile() <= file->GetSourceCount())
			return;
    
		CSourceData* cur_src = sources->GetAt(pos);
		CUpDownClient* newclient = new CUpDownClient(cur_src->sourcePort, cur_src->sourceID, 0, 0, file);
	 	theApp.downloadqueue->CheckAndAddSource(file, newclient);
        
	}
	//DEBUG_ONLY(theApp.emuledlg->AddLogLine(/*TBN_NONOTIFY, */false, "Loaded %i sources for file %s", sources->GetCount(), file->GetFileName()));	

}

//#define SOURCESTOSAVE	10
#define EXPIREIN		 3 //days

void CSourceSaver::SaveSources(CPartFile* file, SourceList* prevsources, CString& slsfile, int maxSourcesToSave)
{
	SourceList srcstosave;
	CSourceData* sourcedata;

	ASSERT(srcstosave.IsEmpty());

	POSITION pos, pos2;

	// Choose best sources for the file
	for (int sl=0;sl<SOURCESSLOTS;sl++) 
	{
		if (!file->srclists[sl].IsEmpty()) 
		{
			for (pos = file->srclists[sl].GetHeadPosition();pos != 0;file->srclists[sl].GetNext(pos)){
				CUpDownClient* cur_src = file->srclists[sl].GetAt(pos);
				if (cur_src->HasLowID())
					continue;
				if (srcstosave.IsEmpty()) {
					sourcedata = new CSourceData(cur_src, CalcExpiration(EXPIREIN));
					srcstosave.AddHead(sourcedata);
					continue;
				}
				if ((srcstosave.GetCount() < maxSourcesToSave) || (cur_src->GetAvailablePartCount() > srcstosave.GetTail()->partsavailable)) {
					if (srcstosave.GetCount() == maxSourcesToSave)
						delete srcstosave.RemoveTail();
					ASSERT(srcstosave.GetCount() < maxSourcesToSave);
					bool bInserted = false;
					for (pos2 = srcstosave.GetTailPosition();pos2 != 0;srcstosave.GetPrev(pos2)){
						CSourceData* cur_srctosave = srcstosave.GetAt(pos2);
						if (cur_srctosave->partsavailable > cur_src->GetAvailablePartCount()) {
							sourcedata = new CSourceData(cur_src, CalcExpiration(EXPIREIN));
							srcstosave.InsertAfter(pos2, sourcedata);
							bInserted = true;
							break;
						}
					}
					if (!bInserted) {
						sourcedata = new CSourceData(cur_src, CalcExpiration(EXPIREIN));
						srcstosave.AddHead(sourcedata);
					}
				}
			}	
		}
	}
	
	// Add previously saved sources if found sources does not reach the limit
	for (pos = prevsources->GetHeadPosition(); pos; prevsources->GetNext(pos)) {
		CSourceData* cur_sourcedata = prevsources->GetAt(pos);
		if (srcstosave.GetCount() == maxSourcesToSave)
			break;
		ASSERT(srcstosave.GetCount() <= maxSourcesToSave);

		bool bFound = false;
		for (pos2 = srcstosave.GetHeadPosition(); pos2; srcstosave.GetNext(pos2)) {
			if (srcstosave.GetAt(pos2)->Compare(cur_sourcedata)) {
				bFound = true;
				break;
			}
		}
		if (!bFound) {
			srcstosave.AddTail(new CSourceData(cur_sourcedata));
		}
			
	}

	//DEBUG_ONLY(theApp.emuledlg->AddLogLine(/*TBN_NONOTIFY, */false, "Saving %i sources for file %s", srcstosave.GetCount(), file->GetFileName()));	

	CString strLine;
	CStdioFile f;
	if (!f.Open(slsfile, CFile::modeCreate | CFile::modeWrite | CFile::typeText))
		return;
	f.WriteString("#format: a.b.c.d:port,expirationdate(yymmdd);\r\n");
	while (!srcstosave.IsEmpty()) {
		CSourceData* cur_src = srcstosave.RemoveHead();
		uint32 dwID = cur_src->sourceID;
		uint16 wPort = cur_src->sourcePort;
		strLine.Format("%i.%i.%i.%i:%i,%s;\r\n", (uint8)dwID,(uint8)(dwID>>8),(uint8)(dwID>>16),(uint8)(dwID>>24), wPort, cur_src->expiration);
		delete cur_src;
		f.WriteString(strLine);
	}
	f.Close();
}

CString CSourceSaver::CalcExpiration(int nDays)
{
	CTime expiration = CTime::GetCurrentTime();
	CTimeSpan timediff(nDays, 0, 0, 0);
	expiration += timediff;
    
	CString strExpiration;
	strExpiration.Format("%02i%02i%02i", (expiration.GetYear() % 100), expiration.GetMonth(), expiration.GetDay());

	return strExpiration;
}

bool CSourceSaver::IsExpired(CString expirationdate)
{
	int year = atoi(expirationdate.Mid(0, 2)) + 2000;
	int month = atoi(expirationdate.Mid(2, 2));
	int day = atoi(expirationdate.Mid(4, 2));

	CTime expiration(year, month, day, 0, 0, 0);
	return (expiration < CTime::GetCurrentTime());
}
