#include "stdafx.h"
#include "KnownFile.h"
#include "KnownFileList.h"





//--- xrmb > partTraffic ---
//--- file > KnownFile.cpp ---

//-------------------------------------------------------------------------------------------------
CFileStatistic::CFileStatistic()
{
	partTraffic=NULL;
	partTrafficSession=NULL;

	blockTraffic=NULL;
	blockTrafficSession=NULL;

	partAccepted=NULL;
	partAcceptedSession=NULL;

	completeReleases=0.0;
	requested = transfered = accepted = alltimerequested= alltimetransfered = alltimeaccepted = 0;
}



//-------------------------------------------------------------------------------------------------
CFileStatistic::~CFileStatistic()	
{
	if(partTraffic)				delete[] partTraffic;			partTraffic=NULL;
	if(partTrafficSession)		delete[] partTrafficSession;	partTrafficSession=NULL;

	if(partAccepted)			delete[] partAccepted;			partAccepted=NULL;
	if(partAcceptedSession)		delete[] partAcceptedSession;	partAcceptedSession=NULL;

	if(blockTraffic)			delete[] blockTraffic;			blockTraffic=NULL;
	if(blockTrafficSession)		delete[] blockTrafficSession;	blockTrafficSession=NULL;
}



//-------------------------------------------------------------------------------------------------
void CFileStatistic::initTraffic()
{
	if(partTraffic==NULL)
	{
		uint16 p=fileParent->GetPartCount();
		partTraffic=new uint32[p];
		partTrafficSession=new uint32[p];
		for(uint16 i=0; i<p; i++)
		{
			partTraffic[i]=0;
			partTrafficSession[i]=0;
		}
	}

	if(partAccepted==NULL)
	{
		uint16 p=fileParent->GetPartCount();
		partAccepted=new uint32[p];
		partAcceptedSession=new uint32[p];
		for(uint16 i=0; i<p; i++)
		{
			partAccepted[i]=0;
			partAcceptedSession[i]=0;
		}
	}

	if(blockTraffic==NULL)
	{
		uint32 p=fileParent->GetBlockCount();
		blockTraffic=new uint32[p];
		blockTrafficSession=new uint32[p];
		for(uint32 i=0; i<p; i++)
		{
			blockTraffic[i]=0;
			blockTrafficSession[i]=0;
		}
	}
}



//-------------------------------------------------------------------------------------------------
void CFileStatistic::recalcCompleteReleases()
{
	//--- recalculate complete releases factor ---
	uint32 blocks=fileParent->GetBlockCount();
	uint32 fs=fileParent->GetFileSize();

	//--- find lowest parttraffic ---
	uint32 m=0xFFFFFFFF;
	completeReleases=0;
	for(uint32 i=0; i<blocks; i++)
	{
		if(blockTraffic[i]==0)
			continue;

		uint32 s=BLOCKSIZE;
		if(BLOCKSIZE*(i+1) > fs) 
			s = fs % BLOCKSIZE; 
		
		if(s)
		{
			uint32 f=(uint32)(double)blockTraffic[i]/s+(blockTraffic[i]==s?0:1);
			if(f<m) m=f;
		}
	}

	//--- get avg. ---
	for(uint32 i=0; i<blocks; i++)
	{
		if(blockTraffic[i]==0)
			continue;

		uint32 s=BLOCKSIZE;
		if(BLOCKSIZE*(i+1) > fs) 
			s = fs % BLOCKSIZE;

		if(s)
		{
			double f=(double)blockTraffic[i]/s;
			if(f>m) f=m;
			completeReleases+=f;
		}
	}

	completeReleases/=blocks;
}



//-------------------------------------------------------------------------------------------------
//--- this function may not return a 100% accurate value, because you can't fill a part with blocks
double CFileStatistic::GetCompletePartReleases(uint16 part)
{
	//--- check part, damn that all needs time, why we just cant trust the devs :) ---
	if(partTraffic==NULL || part<0 || part>fileParent->GetPartCount() || partTraffic[part]==0)
		return 0;

	//--- start and end block ---
	uint32	startBlock=(part*PARTSIZE)/BLOCKSIZE;
	uint32	endBlock=((part+1)*PARTSIZE)/BLOCKSIZE;
	if(endBlock >= fileParent->GetBlockCount())
		endBlock=fileParent->GetBlockCount()-1;

	uint32	fs=fileParent->GetFileSize();

	//--- find lowest parttraffic ---
	uint32	m=0xFFFFFFFF;
	double	completePartReleases=0;
	for(uint32 i=startBlock; i<=endBlock; i++)
	{
		if(blockTraffic[i]==0)
			continue;

		uint32 s=BLOCKSIZE;
		if(BLOCKSIZE*(i+1) > fs) 
			s = fs % BLOCKSIZE; 
		
		if(s)
		{
			uint32 f=(uint32)(double)blockTraffic[i]/s+(blockTraffic[i]==s?0:1);
			if(f<m) m=f;
		}
	}

	//--- get avg. ---
	for(uint32 i=startBlock; i<=endBlock; i++)
	{
		if(blockTraffic[i]==0)
			continue;

		uint32 s=BLOCKSIZE;
		if(BLOCKSIZE*(i+1) > fs) 
			s = fs % BLOCKSIZE;

		if(s)
		{
			double f=(double)blockTraffic[i]/s;
			if(f>m) f=m;
			completePartReleases+=f;
		}
	}

	return completePartReleases/(fileParent->GetPartSize(part)/BLOCKSIZE);
}



//-------------------------------------------------------------------------------------------------
void CFileStatistic::AddTraffic(uint32 start, uint32 bytes)
{
	initTraffic();

	//DEBUG_ONLY(theApp.emuledlg->AddLogLine(false, "--------------------------------------------------------------------------------------------------"));
	//DEBUG_ONLY(theApp.emuledlg->AddLogLine(false, "start %i, bytes %i ", start, bytes));

	//--- start block ---
	uint32	part=start/PARTSIZE;
	sint32	btg=(sint32)bytes;

	uint32 p1=min((part+1)*PARTSIZE-start, bytes);
	partTraffic[part]+=p1;
	partTrafficSession[part]+=p1;
	partAccepted[part]++;
	partAcceptedSession[part]++;
	btg-=p1;
	//DEBUG_ONLY(theApp.emuledlg->AddLogLine(false, "add %i bytes to part %i, %i bytes left", p1, part, btg));
	part++;
	while(btg>0)
	{
		//DEBUG_ONLY(theApp.emuledlg->AddLogLine(false, "add %i bytes to part %i, %i bytes left", btg % PARTSIZE, part, max(btg-PARTSIZE, 0)));
		partTraffic[part]+=btg % PARTSIZE;
		partTrafficSession[part]+=btg % PARTSIZE;
		partAccepted[part]++;
		partAcceptedSession[part]++;
		part++;
		btg-=PARTSIZE;
	}

	//--- start block ---
	uint32	block=start/BLOCKSIZE;
	btg=(sint32)bytes;

	uint32 b1=min((block+1)*BLOCKSIZE-start, bytes);
	blockTraffic[block]+=b1;
	blockTrafficSession[block]+=b1;
	btg-=b1;
	//DEBUG_ONLY(theApp.emuledlg->AddLogLine(false, "add %i bytes to block %i, %i bytes left", b1, block, btg));
	block++;
	while(btg>0)
	{
		//DEBUG_ONLY(theApp.emuledlg->AddLogLine(false, "add %i bytes to block %i, %i bytes left", btg % BLOCKSIZE, block, max(btg-BLOCKSIZE, 0)));
		blockTraffic[block]+=btg % BLOCKSIZE;
		blockTrafficSession[block]+=btg % BLOCKSIZE;
		block++;
		btg-=BLOCKSIZE;
	}

	recalcCompleteReleases();
	
	theApp.sharedfiles->UpdateItem(fileParent);
}



//-------------------------------------------------------------------------------------------------
bool	CFileStatistic::merge(CFileStatistic *m)
{
	alltimerequested	+= m->alltimerequested;
	alltimeaccepted		+= m->alltimeaccepted;
	alltimetransfered	+= m->alltimetransfered;

	requested	+= m->requested;
	accepted	+= m->accepted;
	transfered	+= m->transfered;

	//--- merge parttraffic ---

	//--- i dont have traffic, but the to merge has ---
	if(partTraffic==NULL && m->partTraffic)
		initTraffic();

	//--- merge traffic only if the to merge has some ---
	if(m->partTraffic)
	{
		for(uint16 part=0; part<fileParent->GetPartCount(); part++)
			partTraffic[part]=m->partTraffic[part];

		for(uint32 block=0; block<fileParent->GetBlockCount(); block++)
			blockTraffic[block]+=m->blockTraffic[block];
	}

	return true;
}



//-------------------------------------------------------------------------------------------------
uint8	CKnownFile::GetPartStatus(uint16 part)
{ 
	uint8 r=PR_PART_ON; 
	if(m_partStatus.Lookup(part, r)) 
		return r; 
	else 
		return PR_PART_ON; 
}


//-------------------------------------------------------------------------------------------------
bool	CKnownFile::HasHiddenParts()
{
	POSITION	pos=m_partStatus.GetStartPosition();
	uint16		part;
	uint8		status;
	while(pos)
	{
		m_partStatus.GetNextAssoc(pos, part, status);
		if(status!=PR_PART_ON)
			return true;
	}
	return false;
}



//-------------------------------------------------------------------------------------------------
void CKnownFile::WritePartStatus(CFile* file)
{
	uint16 parts = hashlist.GetCount();
	file->Write(&parts, 2);
	uint16 done = 0;
	while(done!=parts)
	{
		uint8 towrite = 0;
		for (uint32 i = 0;i != 8;i++){
			if (GetPartStatus(done)==PR_PART_ON)
				towrite |= (1<<i);

			done++;
			if (done == parts)
				break;
		}
		file->Write(&towrite,1);
	}
}



//-------------------------------------------------------------------------------------------------
bool CKnownFile::SaveToFileTraffic(FILE* file)
{
	if(statistic.partTraffic)
	{
		uint16 parts=GetPartCount();
		if(fwrite(&parts, sizeof(parts), 1, file)!=1) return false;
		if(fwrite(statistic.partTraffic, 4, parts, file)!=parts) return false;

		uint32 blocks=GetBlockCount();
		if(fwrite(&blocks, sizeof(blocks), 1, file)!=1) return false;
		if(fwrite(statistic.blockTraffic, 4, blocks, file)!=blocks) return false;

		//--- added in version 2 ---
		if(fwrite(statistic.partAccepted, 4, parts, file)!=parts) return false;
	}

	//--- this block shouldnt be used at all ---
	else
	{
		uint32	d=0;

		uint16 parts=GetPartCount();
		if(fwrite(&parts, sizeof(parts), 1, file)!=1) return false;
		for(uint16 part=0; part<parts; part++)
			if(fwrite(&d, sizeof(d), 1, file)!=1) return false;

		uint32 blocks=GetBlockCount();
		if(fwrite(&blocks, sizeof(blocks), 1, file)!=1) return false;
		for(uint32 block=0; block<blocks; block++)
			if(fwrite(&d, sizeof(d), 1, file)!=1) return false;

		//--- added version 2 ---
		for(uint16 part=0; part<parts; part++)
			if(fwrite(&d, sizeof(d), 1, file)!=1) return false;
	}

	return ferror(file);
}



//-------------------------------------------------------------------------------------------------
bool CKnownFile::LoadFromFileTraffic(FILE* file, uint8 version)
{
	// read parts
	uint16 parts=0;
	if(fread(&parts, sizeof(parts), 1, file)!=1) return false;
	
	// does part count match?
	if(parts!=GetPartCount())
	{
		theApp.emuledlg->AddLogLine(true, "traffic.dat is corrupted");
		return false;
	}

	// get the memory
	statistic.initTraffic();

	// load all the parts traffic data
	if(fread(statistic.partTraffic, 4, parts, file)!=parts) return false;

	// read blocks
	uint32 blocks=0;
	if(fread(&blocks, sizeof(blocks), 1, file)!=1) return false;
	
	// does blocks count match?
	if(blocks!=GetBlockCount())
	{
		theApp.emuledlg->AddLogLine(true, "traffic.dat is corrupted");
		return false;
	}

	// load all the parts traffic data
	if(fread(statistic.blockTraffic, 4, blocks, file)!=blocks) return false;

	//--- versioning ---
	if(version>1)
	{
		//--- version 2 introduced part-accepted counter ---
		if(fread(statistic.partAccepted, 4, parts, file)!=parts) return false;
	}

	//--- recalc something ---
	statistic.recalcCompleteReleases();

	//--- parts repair/recalc ---
	/*
	if(true)
	{
		for(uint16 part=0; part<parts; part++)
			statistic.partTraffic[part]=statistic.fileParent->GetTrafficBlock(part*PARTSIZE, part*PARTSIZE+PARTSIZE);
	}
	*/
	return true;
}



//-------------------------------------------------------------------------------------------------
uint32 CKnownFile::GetBlockCount()
{
	return (m_nFileSize/BLOCKSIZE) + ((m_nFileSize % BLOCKSIZE)? 1:0);
}



//-------------------------------------------------------------------------------------------------
uint32 CKnownFile::GetTrafficBlock(uint32 start, uint32 end, bool session)
{
	if(statistic.blockTraffic==NULL)
		return 0;

	if(start>m_nFileSize) return 0;
	if(end>m_nFileSize) end=m_nFileSize;

	uint32	*data;
	if(session)
		data=statistic.blockTrafficSession;
	else
		data=statistic.blockTraffic;

	uint32	block=start/BLOCKSIZE;
	uint32	pos=start;
	sint32	bftb;
	uint32	t=0;
	while(pos < end)
	{
		if(end-pos > end % BLOCKSIZE)
			bftb=BLOCKSIZE;
		else
			bftb=end % BLOCKSIZE;
		bftb-=pos % BLOCKSIZE;

		if(bftb != BLOCKSIZE)
			t+=data[block] * (double)bftb/GetBlockSize(block);
		else
			t+=data[block];

		pos+=bftb;

		block++;
	}

	return t;
}



//-------------------------------------------------------------------------------------------------
uint32 CKnownFile::GetTrafficPart(uint32 start, uint32 end, bool session)
{
	if(statistic.partTraffic==NULL)
		return 0;

	if(start>m_nFileSize) return 0;
	if(end>m_nFileSize) end=m_nFileSize;

	uint32	*data;
	if(session)
		data=statistic.partTrafficSession;
	else
		data=statistic.partTraffic;

	uint32	part=start/PARTSIZE;
	uint32	pos=start;
	sint32	bftp;
	uint32	t=0;
	while(pos < end)
	{
		if(end-pos > end % PARTSIZE)
			bftp=PARTSIZE;
		else
			bftp=end % PARTSIZE;
		bftp-=pos % PARTSIZE;

		if(bftp != PARTSIZE)
			t+=data[part] * (double)bftp/GetPartSize(part);
		else
			t+=data[part];

		pos+=bftp;

		part++;
	}

	return t;
}



//-------------------------------------------------------------------------------------------------
uint32 CKnownFile::GetPartTraffic(uint16 part, bool session)
{
	if(statistic.partTraffic==NULL)
		return 0;

	if(session)
		return statistic.partTrafficSession[part];
	else
		return statistic.partTraffic[part];
}



//-------------------------------------------------------------------------------------------------
uint32 CKnownFile::GetBlockTraffic(uint16 block, bool session)
{
	if(statistic.partTraffic==NULL)
		return 0;

	if(session)
		return statistic.blockTrafficSession[block];
	else
		return statistic.blockTraffic[block];
}



//-------------------------------------------------------------------------------------------------
uint32 CKnownFile::GetPartSize(uint16 part)
{
	if(part+1<GetPartCount())
		return PARTSIZE;

	return m_nFileSize % PARTSIZE;
}



//-------------------------------------------------------------------------------------------------
uint32 CKnownFile::GetBlockSize(uint32 block)
{
	if(block+1<GetBlockCount())
		return BLOCKSIZE;

	return m_nFileSize % BLOCKSIZE;
}



//-------------------------------------------------------------------------------------------------
void CFileStatistic::resetStats(bool all)
{
	if(all)
	{
		if(partTraffic)		delete[] partTraffic;	partTraffic=NULL;
		if(partAccepted)	delete[] partAccepted;	partAccepted=NULL;
		if(blockTraffic)	delete[] blockTraffic;	blockTraffic=NULL;
	
		if(partTrafficSession)		delete[] partTrafficSession;	partTrafficSession=NULL;
		if(partAcceptedSession)		delete[] partAcceptedSession;	partAcceptedSession=NULL;
		if(blockTrafficSession)		delete[] blockTrafficSession;	blockTrafficSession=NULL;

		completeReleases=0.0;

		alltimerequested=alltimetransfered=alltimeaccepted=0;
	}
	else
	{
		if(partTrafficSession)
			for(uint16 part=0; part<fileParent->GetPartCount(); part++)
				partTrafficSession[part]=0;

		if(partAcceptedSession)
			for(uint16 part=0; part<fileParent->GetPartCount(); part++)
				partAcceptedSession[part]=0;

		if(blockTrafficSession)
			for(uint32 block=0; block<fileParent->GetBlockCount(); block++)
				blockTrafficSession[block]=0;
	}
	
	requested=transfered=accepted=0;
}



//-------------------------------------------------------------------------------------------------
void CFileStatistic::resetPartTraffic(uint16 part, bool all)
{
	if(partTraffic==NULL || part>=fileParent->GetPartCount())
		return;

	//--- this all is a little bit tricky, because the part and block size have no connections ---
	//--- so i have to remove some bytes only from the start and end block ---
	//--- all we try is to keep it as realistic as possible :) ---
	uint32	blocks=fileParent->GetBlockCount();
	uint32	blockStart=(part*PARTSIZE)/BLOCKSIZE;
	uint32	blockEnd=min((((uint32)part+1)*PARTSIZE)/BLOCKSIZE+1, blocks-1);
	float	percentStart=((float)((part*PARTSIZE)%BLOCKSIZE)/BLOCKSIZE);
	float	percentEnd=0.0;
	if(blockEnd!=blocks)
		percentEnd=1-((float)(((part+1)*PARTSIZE)%BLOCKSIZE)/BLOCKSIZE);

	if(all)
	{
		partTraffic[part]=0;
		blockTraffic[blockStart]*=percentStart;
		blockTraffic[blockEnd]*=percentEnd;
		for(uint32 block=blockStart+1; block<blockEnd; block++)
			blockTraffic[block]=0;
	}

	partTrafficSession[part]=0;
	blockTrafficSession[blockStart]*=percentStart;
	blockTrafficSession[blockEnd]*=percentEnd;
	for(uint32 block=blockStart+1; block<blockEnd; block++)
		blockTrafficSession[block]=0;
}



//-------------------------------------------------------------------------------------------------
uint32	CFileStatistic::GetPartAccepted(uint16 part, bool session) const
{
	if(partTraffic==NULL || part>=fileParent->GetPartCount())
		return 0;

	if(session==false)
		return partAccepted[part];
	else
		return partAcceptedSession[part];
}



//-------------------------------------------------------------------------------------------------
bool CKnownFile::ExportParts(const CList<uint16> &parts)
{
	if(parts.GetCount()==0)
		return false;

	CPartFile *cpf=NULL;
	if(IsPartFile())
		cpf=(CPartFile*)this;

	POSITION	pos;
	uint16		part;
	if(cpf)
	{
		pos=parts.GetHeadPosition();
		while(pos)
		{
			part=parts.GetNext(pos);

			//--- are all parts complete? ---
			if(cpf->IsComplete(PARTSIZE*part, PARTSIZE*part+PARTSIZE-1)==false)
			{
				theApp.emuledlg->AddLogLine(true, "Can't export incomplete part %i of file '%s'...", part, GetFileName());
				return false;
			}
		}
	}

	//--- default filename is part hash (if only one part...) ---
	char buffer[MAX_PATH];
	if(parts.GetCount()==1)
	{
		sprintf(buffer, "part_%03i", parts.GetHead());
	}
	else
	{
		pos=parts.GetHeadPosition();
		sprintf(buffer, "parts_");
		while(pos)
		{
			part=parts.GetNext(pos);
			sprintf(buffer+strlen(buffer), "%03i%c", part, pos?',':'\0');

			if(strlen(buffer)>100) 
			{
				strcat(buffer, "...");
				break;
			}
		}
	}
	sprintf(buffer+strlen(buffer), "_of_");
	for(uint16 i=0; i!=16; i++)
		sprintf(buffer+strlen(buffer), "%02X", GetFileHash()[i]);
	strcat(buffer, ".part");

	CFileDialog dlg(false, "part", buffer, OFN_HIDEREADONLY);

	//--- try to read the part ---
	if(dlg.DoModal()!=IDOK)
	{
		return false;
	}

	CString pathName=dlg.GetPathName();

	//--- file exists? ---
	if(PathFileExists(pathName))
	{
		if(IDNO==MessageBox(NULL, "Overwrite file '"+pathName+"'?",  "Confirm", MB_YESNO)) 
			return false;
	}

	FILE	*fh=fopen(pathName, "wb");
	if(fh==NULL)
	{
		theApp.emuledlg->AddLogLine(true, "Couldn't create file '%s'...", pathName.GetBuffer());
		return false;
	}

	//--- this is great, isnt it? ---
	uchar	*partData=new uchar[PARTSIZE];
	uint32	partSize;

	pos=parts.GetHeadPosition();
	while(pos)
	{
		part=parts.GetNext(pos);
		partSize=GetPartSize(part);

		//--- try to read the part ---
		if(cpf)
		{
			//--- get pos ---
			ULONGLONG	pos=cpf->m_hpartfile.GetPosition();

			cpf->m_hpartfile.Seek(PARTSIZE*part, CFile::begin);

			//--- read da part ---
			bool success=(cpf->m_hpartfile.Read(partData, partSize)==partSize);

			//--- restore pos ---
			cpf->m_hpartfile.Seek(pos, CFile::begin);

			if(success==false)
			{
				theApp.emuledlg->AddLogLine(true, "Couldn't read part %i file '%s'...", part, GetFileName());
				fclose(fh);
				unlink(pathName);
				delete partData;
				return false;
			}
		}
		else
		{
			char	buffer[MAX_PATH];
			sprintf(buffer, "%s\\%s", GetPath(), GetFileName());
			CFile	f;
			if(f.Open(buffer, CFile::modeRead)==false)
			{
				theApp.emuledlg->AddLogLine(true, "Couldn't open file '%s'...", buffer);
				fclose(fh);
				unlink(pathName);
				delete partData;
				return false;
			}

			f.Seek(PARTSIZE*part, CFile::begin);

			bool success=(f.Read(partData, partSize)==partSize);

			f.Close();

			if(success==false)
			{
				theApp.emuledlg->AddLogLine(true, "Couldn't read part %i file '%s'...", part, GetFileName());
				fclose(fh);
				unlink(pathName);
				delete partData;
				return false;
			}
		}

		//--- no check if write was ok!!! ---
		if(fwrite(partData, partSize, 1, fh)!=1)
		{
			theApp.emuledlg->AddLogLine(true, "Couldn't write into file '%s'...", pathName.GetBuffer());
			fclose(fh);
			unlink(pathName);
			delete partData;
			return false;
		}

		theApp.emuledlg->AddLogLine(true, "part %i successfully exported...", part);
	}	

	fclose(fh);
	delete partData;
	return true;
}



//-------------------------------------------------------------------------------------------------
bool CKnownFile::ImportParts()
{
	//--- import to a complete file? forget it! ---
	if(IsPartFile()==false)
	{
		theApp.emuledlg->AddLogLine(true, "Can't import parts to a completed file...");
		return false;
	}

	CPartFile* cpf=(CPartFile*)this;

	char buffer[MAX_PATH];
	strcpy(buffer, "*.part");

	CFileDialog dlg(true, "part", buffer, OFN_FILEMUSTEXIST|OFN_HIDEREADONLY);

	//--- try to read the part ---
	if(dlg.DoModal()!=IDOK)
		return false;

	CString pathName=dlg.GetFileName();

	//--- read part content ---
	CFile	f;
	if(f.Open(pathName, CFile::modeRead)==false)
	{
		theApp.emuledlg->AddLogLine(true, "Couldn't open file '%s'...", pathName.GetBuffer());
		return false;
	}

	//--- read the data ---
	uchar	*partData=new uchar[PARTSIZE];
	uint32	partSize;
	uint16	ipart=-1, part;
	do
	{
		partSize=f.Read(partData, PARTSIZE);
		ipart++;

		if(partSize==0)
			break;

		//--- calc hash for part ---
		uchar	hash[16];
		CreateHashFromInput(0, 0, partSize, hash, partData);

		buffer[0]=0;
		for(uint16 i=0; i!=16; i++)
			sprintf(buffer+strlen(buffer), "%02X", hash[i]);

		bool found=false;
		for(uint16 i=0; i<GetPartCount(); i++)
			if(memcmp(hash, cpf->GetPartHash(i), 16)==0)
			{
				part=i;
				found=true;
				theApp.emuledlg->AddLogLine(true, "A match for part %i was found...", part);
				break;
			}

		if(found==false)
		{
			theApp.emuledlg->AddLogLine(true, "The imported part %i is not a part of this file.", ipart);
			continue;
		}

		//--- overwrite a completed part? ---
		if(cpf->IsComplete(part*PARTSIZE, part*PARTSIZE+PARTSIZE-1))
		{
			char buffer[100];
			sprintf(buffer, "Import already completed part %i?", part);
			if(IDNO==MessageBox(NULL, buffer, "Confirm", MB_YESNO)) 
			{
				theApp.emuledlg->AddLogLine(true, "Part %i skipped, is already complete...", part);
				continue;
			}
		}
		else
		{
			char buffer[100];
			sprintf(buffer, "The part %i seems to fit in this file, import it?", part);
			if(IDNO==MessageBox(NULL, buffer, "Confirm", MB_YESNO)) 
			{
				theApp.emuledlg->AddLogLine(true, "Part %i skipped, you said no...", part);
				continue;
			}
		}

		//--- for partfiles the import is cheating the system :) ---
		//--- we read the whole part, and send it to the "BlockReceived" ---
		// TODO, argh they changeds the receive function :(
		//cpf->BlockReceived(part*PARTSIZE, part*PARTSIZE+partSize, (char*)partData);

		theApp.emuledlg->AddLogLine(true, "importing part %i successfully done...", part);
	}
	while(partSize==PARTSIZE);

	f.Close();
	delete partData;
	return true;
}

//--- file < KnownFile.cpp ---




//--- file > KnownFileList.cpp ---
//-------------------------------------------------------------------------------------------------
bool	CKnownFileList::LoadPartTraffic()
{
	//--- load collected traffic data in our file ---
	char	fullpath[MAX_PATH];
	strcpy(fullpath, appdir);
	strcat(fullpath, "traffic.dat");

	FILE	*fh=fopen(fullpath, "rb");
	if(!fh)
		return false;

	uint8	version=0;
	fread(&version, sizeof(version), 1, fh);

	uint32	records;
	fread(&records, sizeof(records), 1, fh);

	uchar tmp_filehash[16];
	uint32 fs=0;
	for(uint32 i=0; i<records; i++) 
	{
		uint8 hasData=0;
		fread(&hasData, sizeof(hasData), 1, fh);

		if(hasData)
		{
			fread(tmp_filehash, 16, 1, fh);
			fread(&fs, sizeof(fs), 1, fh);
			if(fs==ElementAt(i)->GetFileSize() && memcmp(tmp_filehash, ElementAt(i)->GetFileHash(), 16)==0)
				ElementAt(i)->LoadFromFileTraffic(fh, version);
			else
			{
				theApp.emuledlg->AddLogLine(false, "'traffic.dat' and 'known.met' aren't synchronous after %d datasets.", i);
				break;
			}
		}
	}
	fclose(fh);

	return true;
}



//-------------------------------------------------------------------------------------------------
bool	CKnownFileList::SavePartTraffic()
{
	//--- save traffic data to our file ---
	char fullpath[MAX_PATH];
	strcpy(fullpath, appdir);
	strcat(fullpath, "traffic.dat");

	FILE	*file=fopen(fullpath, "wb");
	if(file==NULL)
		return false;

	uint8	version=2;
	fwrite(&version, sizeof(version), 1, file);	

	uint32	RecordsNumber=GetCount();
	fwrite(&RecordsNumber, sizeof(RecordsNumber), 1, file);
	for(uint32 i=0; i<RecordsNumber; i++) 
	{
		// only save for known-files with traffic
		if(ElementAt(i)->statistic.partTraffic)
		{
			uint8 hasData=1;
			fwrite(&hasData, sizeof(hasData), 1, file);

			//--- write hash ---
			fwrite(ElementAt(i)->GetFileHash(), 16, 1, file);

			//--- write filesize ---
			uint32 fs=ElementAt(i)->GetFileSize();
			fwrite(&fs, sizeof(fs), 1, file);
			
			//--- write part/block traffic ---
			ElementAt(i)->SaveToFileTraffic(file);
		}
		else
		{
			uint8 hasData=0;
			fwrite(&hasData, sizeof(hasData), 1, file);
		}
	}

	fclose(file);

	return true;
}



//-------------------------------------------------------------------------------------------------
bool	CKnownFileList::LoadPartPrio()
{
	//--- save the part-permission ---
	char	fullpath[MAX_PATH];
	strcpy(fullpath, appdir);
	strcat(fullpath, "partperm.dat");

	FILE	*file=fopen(fullpath, "rb");
	if(file==NULL)
		return false;

	uint8 version=1;
	fread(&version, sizeof(version), 1, file);	

	uint32	RecordsNumber=0;
	fread(&RecordsNumber, sizeof(RecordsNumber), 1, file);

	CKnownFile	*kf;
	uchar		hash[16];
	uint32		filesize;
	uint32		fpos;
	for(uint32 i=0; i<RecordsNumber; i++) 
	{
		//--- read hash ---
		fread(hash, 16, 1, file);

		//--- read filesize ---
		fread(&filesize, sizeof(filesize), 1, file);

		//--- save pos ---
		fpos=ftell(file);
		
		//--- find all these files ---
		for(int j=0; j<GetCount(); j++)
		{
			kf=ElementAt(j);

			//--- right file? ---
			if(kf->GetFileSize()==filesize && memcmp(kf->GetFileHash(), hash, 16)==0)
			{
				fseek(file, fpos, SEEK_SET);

				for(uint16 part=0; part<kf->GetPartCount(); part++)
				{
					uint8	state=PR_PART_ON;
					fread(&state, sizeof(state), 1, file);
					kf->SetPartStatus(part, state);
				}
			}
		}
	}
	fclose(file);

	return true;
}



//-------------------------------------------------------------------------------------------------
bool	CKnownFileList::SavePartPrio()
{
	//--- save the part-permission ---
	char	fullpath[MAX_PATH];
	strcpy(fullpath, appdir);
	strcat(fullpath, "partperm.dat");

	FILE	*file=fopen(fullpath, "wb");
	if(file==NULL)
		return false;

	uint8 version=1;
	fwrite(&version, sizeof(version), 1, file);	

	//--- preserve the space for the recordnumers ---
	uint32	RecordsNumber=0;
	uint32	rnpos=ftell(file);
	fwrite(&RecordsNumber, sizeof(RecordsNumber), 1, file);
	for(int i=0; i<GetCount(); i++) 
	{
		CKnownFile	*kf=ElementAt(i);

		// only save for known-files with traffic
		if(kf->HasHiddenParts())
		{
			//--- write hash ---
			fwrite(kf->GetFileHash(), 16, 1, file);

			//--- write filesize ---
			uint32 fs=kf->GetFileSize();
			fwrite(&fs, sizeof(fs), 1, file);
			
			//--- save prio ---
			for(uint16 part=0; part<kf->GetPartCount(); part++)
			{
				uint8	state=kf->GetPartStatus(part);
				fwrite(&state, sizeof(state), 1, file);
			}

			RecordsNumber++;
		}
	}

	//--- save the number of stored records ---
	fseek(file, rnpos, SEEK_SET);
	fwrite(&RecordsNumber, sizeof(RecordsNumber), 1, file);

	fclose(file);

	return true;
}



//-------------------------------------------------------------------------------------------------
bool CKnownFileList::merge()
{
	CSingleLock sLock(&list_mut,true);

	int totalcnt=0;
	int items=GetCount();

	//--- go thru all items ---
	for(int i=0; i<items; i++)
	{
		int cnt=0;

		//--- compare with all other items ---
		for(int j=i+1; j<items; j++)
		{
			//--- same id? and same size... maybe we should compare the hashsets too? ---
			if(	ElementAt(i)->GetFileSize()==ElementAt(j)->GetFileSize() &&
				memcmp(ElementAt(i)->GetFileHash(), ElementAt(j)->GetFileHash(), 16)==0
				)
			{
				theApp.emuledlg->AddLogLine(false, "merge: '%s' and '%s'.", ElementAt(i)->GetFileName(), ElementAt(j)->GetFileName());

				//--- one file can be shared ---
				if(ElementAt(i)->GetSharedFile() && ElementAt(j)->GetSharedFile())
				{
					theApp.emuledlg->AddLogLine(false, "merge: you share the same file twice? abort merging.");
					sLock.Unlock();
					return false;
				}

				//--- if second file is shared, swap them ---
				if(ElementAt(j)->GetSharedFile())
				{
					theApp.emuledlg->AddLogLine(false, "merge: swapping files.");

					CKnownFile *t=ElementAt(i);
					SetAt(i, ElementAt(j));
					SetAt(j, t);
				}

				//--- merge into first item ---
				if(ElementAt(i)->statistic.merge(&ElementAt(j)->statistic)==false)
				{
					theApp.emuledlg->AddLogLine(false, "merge: failed.");
					sLock.Unlock();
					return false;
				}

				//--- remove second item ---
				CKnownFile	*d=ElementAt(j);
				RemoveAt(j);
				delete d;
				items--;
				j--;

				cnt++;
				totalcnt++;
			}
		}

		if(cnt)
		{
			theApp.emuledlg->AddLogLine(false, "merge: item %i merged with %i other items.", i, cnt);
		}
	}

	theApp.emuledlg->AddLogLine(false, "merge: %i items merged in total.", totalcnt);

	sLock.Unlock();
	return true;
}



//-------------------------------------------------------------------------------------------------
bool CKnownFileList::RemoveFile(CKnownFile* toRemove)
{
	CSingleLock sLock(&list_mut, true);

	//--- go thru all items ---
	for(int i=0; i<GetCount(); i++)
	{
		if(ElementAt(i)==toRemove)
		{
			RemoveAt(i);
			delete toRemove;
		
			sLock.Unlock();
			return true;
		}
	}

	sLock.Unlock();
	return false;
}



//-------------------------------------------------------------------------------------------------
void CKnownFileList::Process()
{
	if(GetTickCount() - lastsaved > 15*60*1000) // 15min
	{
		lastsaved=GetTickCount();
		Save();
	}
}



//--- file < KnownFileList.cpp ---

//--- xrmb < partTraffic ---



//--- xrmb > keepPartFileStats ---

//--- file > KnownFile.cpp ---

//-------------------------------------------------------------------------------------------------
void CFileStatistic::AddAll(uint32 _alltimerequested, uint32 _alltimeaccepted, uint64 _alltimetransfered)
{
	alltimerequested	+= _alltimerequested;
	alltimeaccepted		+= _alltimeaccepted;
	alltimetransfered	+= _alltimetransfered;
}
//--- file < KnownFile.cpp ---



//--- file > PartFile.cpp ---

//-------------------------------------------------------------------------------------------------
bool CPartFile::LoadPartFileStats()
{
	char	fn[MAX_PATH];
	strcpy(fn, fullname);
	fn[strlen(fn)-4]=0;	// 4 = .met
	strcat(fn, ".stats");
	FILE	*fh=fopen(fn, "rb");
	
	//--- if we cant open this file... you lose the stats ---
	if(fh==NULL)
		return false;
	
	//--- get version ---
	uint8	version=0;
	if(fread(&version, sizeof(version), 1, fh)!=1) { fclose(fh); return false; }

	uint32 alltimerequested=0;
	uint32 alltimeaccepted=0;
	uint64 alltimetransfered=0;

	if(fread(&alltimerequested,		sizeof(alltimerequested),	1, fh)!=1) { fclose(fh); return false; }
	if(fread(&alltimeaccepted,		sizeof(alltimeaccepted),	1, fh)!=1) { fclose(fh); return false; }
	if(fread(&alltimetransfered,	sizeof(alltimetransfered),	1, fh)!=1) { fclose(fh); return false; }

	statistic.AddAll(alltimerequested, alltimeaccepted, alltimetransfered);

	//--- this is for the parttraffic ---
	//--- i thought about explicit adding the traffic like for the other stats above ---
	//--- but it doesnt make sense, because you'll no transfer if you reach this ---
	if(LoadFromFileTraffic(fh, version)==false) { fclose(fh); return false;	}

	fclose(fh);

	return true;
}


//-------------------------------------------------------------------------------------------------
bool CPartFile::SavePartFileStats()
{
	char	fn[MAX_PATH];
	strcpy(fn, fullname);
	fn[strlen(fn)-4]=0;	// 4 = .met
	strcat(fn, ".stats");
	FILE	*fh=fopen(fn, "wb");
	
	//--- if we cant open this file... you lose the stats ---
	if(fh==NULL)
		return false;
	
	//--- set version ---
	uint8	version=2;
	if(fwrite(&version, sizeof(version), 1, fh)!=1) { fclose(fh); return false; }

	uint32 alltimerequested=statistic.GetAllTimeRequests();
	uint32 alltimeaccepted=statistic.GetAllTimeAccepts();
	uint64 alltimetransfered=statistic.GetAllTimeTransfered();

	if(fwrite(&alltimerequested,	sizeof(alltimerequested),	1, fh)!=1) { fclose(fh); return false; }
	if(fwrite(&alltimeaccepted,		sizeof(alltimeaccepted),	1, fh)!=1) { fclose(fh); return false; }
	if(fwrite(&alltimetransfered,	sizeof(alltimetransfered),	1, fh)!=1) { fclose(fh); return false; }

	//--- this is for the parttraffic ---
	if(SaveToFileTraffic(fh)==false) { fclose(fh); return false; }

	fclose(fh);

	return true;
}



//-------------------------------------------------------------------------------------------------
int CPartFile::RemovePartFileStats()
{
	char	fn[MAX_PATH];
	strcpy(fn, fullname);
	fn[strlen(fn)-4]=0;	// 4 = .met
	strcat(fn, ".stats");
	
	return remove(fn);
}

//--- file < partFile.cpp ---



//--- file < DownloadQueue.cpp ---

//-------------------------------------------------------------------------------------------------
void	CDownloadQueue::SaveAllPartFileStats()
{
		for (POSITION pos =filelist.GetHeadPosition();pos != 0;filelist.GetNext(pos))
		{
			CPartFile* cur_file =  filelist.GetAt(pos);
			cur_file->SavePartFileStats();
		}
}
//--- file < DownloadQueue.cpp ---

//--- xrmb < keepPartFileStats ---
